# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bakplane/bakplane.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from bakplane.validate import validate_pb2 as bakplane_dot_validate_dot_validate__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='bakplane/bakplane.proto',
  package='io.openaristos',
  syntax='proto3',
  serialized_options=b'\n\016io.openaristosB\rBakplaneProtoP\001Z\016proto/bakplane\242\002\002BP',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x17\x62\x61kplane/bakplane.proto\x12\x0eio.openaristos\x1a bakplane/validate/validate.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"~\n\x0eTimestampRange\x12\x36\n\x08start_dt\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x08\xfa\x42\x05\xb2\x01\x02\x08\x01\x12\x34\n\x06\x65nd_dt\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x08\xfa\x42\x05\xb2\x01\x02\x08\x01\"3\n\x11PaginationOptions\x12\x10\n\x08per_page\x18\x01 \x01(\r\x12\x0c\n\x04page\x18\x02 \x01(\r\"H\n\x08Universe\x12\x1b\n\ndefinition\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x0f\n\x07\x64ialect\x18\x02 \x01(\t\x12\x0e\n\x06\x65ngine\x18\x03 \x01(\t\"\xb8\x01\n\x0bPerspective\x12\x15\n\x04name\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x19\n\x06weight\x18\x04 \x01(\rB\t\xfa\x42\x06*\x04\x18\x64(\x00\x12K\n\x14permanent_attributes\x18\x05 \x03(\x0b\x32-.io.openaristos.PerspectivePermanentAttribute\"c\n\x1dPerspectivePermanentAttribute\x12!\n\x10\x65ntity_type_code\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x1f\n\x0e\x61ttribute_code\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x83\x02\n\x10MigrationRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x1f\n\x0ewarehouse_kind\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12G\n\tdirection\x18\x03 \x01(\x0e\x32*.io.openaristos.MigrationRequest.DirectionB\x08\xfa\x42\x05\x82\x01\x02\x10\x01\x12\x35\n\x05\x66orce\x18\x04 \x01(\x0b\x32&.io.openaristos.MigrationRequest.Force\x1a\x18\n\x05\x46orce\x12\x0f\n\x07version\x18\x01 \x01(\r\"\x1d\n\tDirection\x12\x06\n\x02UP\x10\x00\x12\x08\n\x04\x44OWN\x10\x01\"P\n\x16MigrationStatusRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x1f\n\x0ewarehouse_kind\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"J\n\x11MigrationResponse\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x16\n\x0ewarehouse_kind\x18\x02 \x01(\t\x12\x0c\n\x04\x63ode\x18\x03 \x01(\t\"\xa2\x02\n\x16ResolveResourceRequest\x12\x34\n\x08universe\x18\x01 \x01(\x0b\x32\x18.io.openaristos.UniverseB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x11\n\twarehouse\x18\x03 \x01(\t\x12\x38\n\x10\x65\x66\x66\x65\x63tive_dating\x18\x04 \x01(\x0b\x32\x1e.io.openaristos.TimestampRange\x12\x46\n\x08pointers\x18\x05 \x03(\x0b\x32\x34.io.openaristos.ResolveResourceRequest.PointersEntry\x1a/\n\rPointersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa2\x02\n\"ResolveResourceMappingContextEntry\x12\x19\n\x11source_field_name\x18\x01 \x01(\t\x12\x1f\n\x17target_field_expression\x18\x02 \x01(\t\x12\x35\n\x0e\x61ttribute_type\x18\x03 \x01(\x0e\x32\x1d.io.openaristos.AttributeType\x12\x18\n\x10used_for_mapping\x18\x04 \x01(\x08\x12\x19\n\x11map_to_field_name\x18\x05 \x01(\t\x12\x1b\n\x13is_effective_dating\x18\x06 \x01(\x08\x12\x37\n\x0f\x63omparison_type\x18\x07 \x01(\x0e\x32\x1e.io.openaristos.ComparisonType\"&\n\x17ResolveResourceResponse\x12\x0b\n\x03row\x18\x01 \x01(\t\"\x84\x03\n\x1dResolveResourceIntentResponse\x12\x16\n\x0ewarehouse_kind\x18\x01 \x01(\t\x12\x12\n\ntable_name\x18\x02 \x01(\t\x12\x13\n\x0bschema_name\x18\x03 \x01(\t\x12=\n\x07mapping\x18\x04 \x01(\x0b\x32,.io.openaristos.ResolveResourceIntentMapping\x12,\n\x07\x63olumns\x18\x05 \x03(\x0b\x32\x1b.io.openaristos.ColumnEntry\x12K\n\x07\x63ontext\x18\x06 \x03(\x0b\x32:.io.openaristos.ResolveResourceIntentResponse.ContextEntry\x12\x38\n\x10\x65\x66\x66\x65\x63tive_dating\x18\x07 \x01(\x0b\x32\x1e.io.openaristos.TimestampRange\x1a.\n\x0c\x43ontextEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xba\x01\n\x1cResolveResourceIntentMapping\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x13\n\x0bschema_name\x18\x02 \x01(\t\x12,\n\x07\x63olumns\x18\x03 \x03(\x0b\x32\x1b.io.openaristos.ColumnEntry\x12\x43\n\x07\x63lauses\x18\x04 \x03(\x0b\x32\x32.io.openaristos.ResolveResourceIntentMappingClause\"J\n\x0b\x43olumnEntry\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61taType\x18\x02 \x01(\t\x12\x1b\n\x13is_effective_dating\x18\x03 \x01(\x08\"\x91\x01\n\"ResolveResourceIntentMappingClause\x12\x1b\n\x13resource_field_name\x18\x01 \x01(\t\x12\x1a\n\x12mapping_field_name\x18\x02 \x01(\t\x12\x32\n\ncomparison\x18\x03 \x01(\x0e\x32\x1e.io.openaristos.ComparisonType\"B\n\x12\x43onfigurationEntry\x12\x14\n\x03key\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01\x12\x16\n\x05value\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01\"|\n\x17\x41pplyLandingZoneRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x0e\n\x06plugin\x18\x02 \x01(\t\x12:\n\x0e\x63onfigurations\x18\x04 \x03(\x0b\x32\".io.openaristos.ConfigurationEntry\"\x1a\n\x18\x41pplyLandingZoneResponse\"2\n\x19RetractLandingZoneRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x1c\n\x1aRetractLandingZoneResponse\"o\n\x16ListLandingZoneRequest\x12\x16\n\x05query\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12=\n\x12pagination_options\x18\x02 \x01(\x0b\x32!.io.openaristos.PaginationOptions\"\xec\x01\n\x17ListLandingZoneResponse\x12Q\n\rlanding_zones\x18\x01 \x03(\x0b\x32:.io.openaristos.ListLandingZoneResponse.LandingZoneDetails\x1a~\n\x12LandingZoneDetails\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x0e\n\x06plugin\x18\x02 \x01(\t\x12\x0e\n\x06\x61\x63tive\x18\x03 \x01(\x08\x12:\n\x0e\x63onfigurations\x18\x04 \x03(\x0b\x32\".io.openaristos.ConfigurationEntry\"1\n\x18\x45nableLandingZoneRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x1b\n\x19\x45nableLandingZoneResponse\"2\n\x19\x44isableLandingZoneRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x1c\n\x1a\x44isableLandingZoneResponse\"\xc8\x02\n\x0fSchemaAttribute\x12\x15\n\x04name\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x1a\n\tdata_type\x18\x04 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01\x12\x31\n\x04mode\x18\x05 \x01(\tB#\xfa\x42 r\x1eR\x08nullableR\x08requiredR\x08repeated\x12\x11\n\tindexable\x18\x06 \x01(\x08\x12K\n\x0cmapping_type\x18\x07 \x01(\x0e\x32+.io.openaristos.SchemaAttribute.MappingTypeB\x08\xfa\x42\x05\x82\x01\x02\x10\x01\"L\n\x0bMappingType\x12\x0e\n\nIDENTIFIER\x10\x00\x12\x16\n\x12PARTIAL_IDENTIFIER\x10\x01\x12\x08\n\x04\x44\x45SC\x10\x02\x12\x0b\n\x07IGNORED\x10\x03\"\x81\x01\n\x06Schema\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12=\n\nattributes\x18\x04 \x03(\x0b\x32\x1f.io.openaristos.SchemaAttributeB\x08\xfa\x42\x05\x92\x01\x02\x08\x01\"x\n\nEntityType\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x30\n\x06schema\x18\x04 \x01(\x0b\x32\x16.io.openaristos.SchemaB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01\"\x96\x01\n\x10RelationshipType\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12#\n\x12source_entity_type\x18\x04 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12#\n\x12target_entity_type\x18\x05 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\"\xc1\x01\n\x14\x41pplyOntologyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x36\n\x08\x65ntities\x18\x04 \x03(\x0b\x32\x1a.io.openaristos.EntityTypeB\x08\xfa\x42\x05\x92\x01\x02\x08\x01\x12\x37\n\rrelationships\x18\x05 \x03(\x0b\x32 .io.openaristos.RelationshipType\"\x17\n\x15\x41pplyOntologyResponse\"f\n\x16RetractOntologyRequest\x12&\n\x15mastering_engine_code\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\r\n\x05\x66orce\x18\x03 \x01(\x08\"\x19\n\x17RetractOntologyResponse\"l\n\x13ListOntologyRequest\x12\x16\n\x05query\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12=\n\x12pagination_options\x18\x02 \x01(\x0b\x32!.io.openaristos.PaginationOptions\"}\n\x14ListOntologyResponse\x12,\n\x08\x65ntities\x18\x01 \x03(\x0b\x32\x1a.io.openaristos.EntityType\x12\x37\n\rrelationships\x18\x02 \x03(\x0b\x32 .io.openaristos.RelationshipType\"K\n\x0f\x45xecutionEngine\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\"\x8e\x01\n\x1b\x41pplyExecutionEngineRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x0c\n\x04kind\x18\x02 \x01(\t\x12\x0e\n\x06plugin\x18\x03 \x01(\t\x12:\n\x0e\x63onfigurations\x18\x04 \x03(\x0b\x32\".io.openaristos.ConfigurationEntry\"\x1e\n\x1c\x41pplyExecutionEngineResponse\"6\n\x1dRetractExecutionEngineRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\" \n\x1eRetractExecutionEngineResponse\"j\n\x1aListExecutionEngineRequest\x12\r\n\x05query\x18\x01 \x01(\t\x12=\n\x12pagination_options\x18\x02 \x01(\x0b\x32!.io.openaristos.PaginationOptions\"O\n\x1bListExecutionEngineResponse\x12\x30\n\x07\x65ngines\x18\x01 \x03(\x0b\x32\x1f.io.openaristos.ExecutionEngine\"5\n\x1c\x45nableExecutionEngineRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x1f\n\x1d\x45nableExecutionEngineResponse\"6\n\x1d\x44isableExecutionEngineRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\" \n\x1e\x44isableExecutionEngineResponse\"E\n\tWarehouse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\"\x88\x01\n\x15\x41pplyWarehouseRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x0e\n\x06plugin\x18\x02 \x01(\t\x12\x0c\n\x04kind\x18\x03 \x01(\t\x12:\n\x0e\x63onfigurations\x18\x04 \x03(\x0b\x32\".io.openaristos.ConfigurationEntry\"\x18\n\x16\x41pplyWarehouseResponse\"0\n\x17RetractWarehouseRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x1a\n\x18RetractWarehouseResponse\"d\n\x14ListWarehouseRequest\x12\r\n\x05query\x18\x01 \x01(\t\x12=\n\x12pagination_options\x18\x02 \x01(\x0b\x32!.io.openaristos.PaginationOptions\"F\n\x15ListWarehouseResponse\x12-\n\nwarehouses\x18\x01 \x03(\x0b\x32\x19.io.openaristos.Warehouse\"/\n\x16\x45nableWarehouseRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x19\n\x17\x45nableWarehouseResponse\"0\n\x17\x44isableWarehouseRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x1a\n\x18\x44isableWarehouseResponse\"\x80\x01\n\x1b\x41pplyMasteringEngineRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x0e\n\x06plugin\x18\x02 \x01(\t\x12:\n\x0e\x63onfigurations\x18\x03 \x03(\x0b\x32\".io.openaristos.ConfigurationEntry\"\x1e\n\x1c\x41pplyMasteringEngineResponse\"6\n\x1dRetractMasteringEngineRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\" \n\x1eRetractMasteringEngineResponse\"7\n\x19ResourceStoreContextEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x86\x01\n\rResourceStore\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x0f\n\x07virtual\x18\x02 \x01(\x08\x12\x11\n\twarehouse\x18\x03 \x01(\t\x12:\n\x07\x63ontext\x18\x04 \x03(\x0b\x32).io.openaristos.ResourceStoreContextEntry\"K\n\x14ResourcePointerEntry\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x1c\n\x0b\x65ntity_type\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"3\n\x15ResourceMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"@\n\x1bResourceContextPointerEntry\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x13\n\x0b\x63olumn_name\x18\x02 \x01(\t\"|\n\x0eResourceDating\x12\x1a\n\x12\x65\x66\x66\x65\x63tive_start_dt\x18\x01 \x01(\t\x12\x18\n\x10\x65\x66\x66\x65\x63tive_end_dt\x18\x02 \x01(\t\x12\x1a\n\x12knowledge_start_dt\x18\x03 \x01(\t\x12\x18\n\x10knowledge_end_dt\x18\x04 \x01(\t\"?\n\x0fResourceHashing\x12\x16\n\x0eknowledge_hash\x18\x01 \x01(\t\x12\x14\n\x0cpayload_hash\x18\x02 \x01(\t\"\xc5\x01\n\x12ResourcePublishing\x12 \n\x18\x64\x64l_create_staging_table\x18\x01 \x01(\t\x12(\n ddl_create_staging_table_indexes\x18\x02 \x01(\t\x12\x1e\n\x16\x64\x64l_drop_staging_table\x18\x03 \x01(\t\x12$\n\x1c\x64ml_obsolete_knowledge_query\x18\x04 \x01(\t\x12\x1d\n\x15\x64ml_incremental_query\x18\x05 \x01(\t\";\n\x12ResourceRetraction\x12\x16\n\x0ewarehouse_kind\x18\x01 \x01(\t\x12\r\n\x05steps\x18\x02 \x03(\t\"\x84\x03\n\x0fResourceContext\x12\x16\n\x0ewarehouse_kind\x18\x01 \x01(\t\x12=\n\x08pointers\x18\x02 \x03(\x0b\x32+.io.openaristos.ResourceContextPointerEntry\x12.\n\x06\x64\x61ting\x18\x03 \x01(\x0b\x32\x1e.io.openaristos.ResourceDating\x12\x30\n\x07hashing\x18\x04 \x01(\x0b\x32\x1f.io.openaristos.ResourceHashing\x12\x37\n\x08metadata\x18\x05 \x03(\x0b\x32%.io.openaristos.ResourceMetadataEntry\x12\x36\n\npublishing\x18\x06 \x01(\x0b\x32\".io.openaristos.ResourcePublishing\x12\x0f\n\x07\x64\x61taset\x18\x07 \x01(\t\x12\x36\n\nretraction\x18\x08 \x01(\x0b\x32\".io.openaristos.ResourceRetraction\"\xb7\x03\n\x08Resource\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x0c\n\x04kind\x18\x02 \x01(\t\x12!\n\x10perspective_code\x18\x03 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x06 \x01(\t\x12\x30\n\x06schema\x18\x07 \x01(\x0b\x32\x16.io.openaristos.SchemaB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01\x12\x36\n\x08pointers\x18\x08 \x03(\x0b\x32$.io.openaristos.ResourcePointerEntry\x12-\n\x06stores\x18\t \x03(\x0b\x32\x1d.io.openaristos.ResourceStore\x12\x17\n\x0fretention_hours\x18\n \x01(\r\x12\x30\n\x07\x63ontext\x18\x0b \x03(\x0b\x32\x1f.io.openaristos.ResourceContext\x12\x37\n\nmigrations\x18\x0c \x03(\x0b\x32\x19.io.openaristos.MigrationB\x08\xfa\x42\x05\x92\x01\x02\x08\x01\x12\x12\n\nis_virtual\x18\r \x01(\x08\"S\n\x07\x41\x64\x61pter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x04 \x01(\t\"f\n\x12\x41\x64\x61pterPersistence\x12\x16\n\x0ewarehouse_kind\x18\x01 \x01(\t\x12\x38\n\x07\x65ntries\x18\x02 \x03(\x0b\x32\'.io.openaristos.AdapterPersistenceEntry\"5\n\x17\x41\x64\x61pterPersistenceEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x82\x01\n\x0f\x41\x64\x61pterMetadata\x12\x37\n\x0bpersistence\x18\x01 \x03(\x0b\x32\".io.openaristos.AdapterPersistence\x12\x36\n\nretraction\x18\x02 \x03(\x0b\x32\".io.openaristos.ResourceRetraction\"d\n\tMigration\x12\x16\n\x0ewarehouse_kind\x18\x01 \x01(\t\x12\x16\n\x0emigrations_uri\x18\x02 \x01(\t\x12\x13\n\x0bschema_name\x18\x03 \x01(\t\x12\x12\n\ntable_name\x18\x04 \x01(\t\"\xfb\x02\n\x13\x41pplyAdapterRequest\x12\x15\n\x04name\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x03\x12\x15\n\x04\x63ode\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x17\n\x06\x61uthor\x18\x04 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x19\n\x08provider\x18\x05 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12L\n\x11\x65xecution_profile\x18\x06 \x01(\x0b\x32\'.io.openaristos.AdapterExecutionProfileB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01\x12\x35\n\nparameters\x18\x07 \x03(\x0b\x32!.io.openaristos.AdapterParameters\x12\x35\n\tresources\x18\x08 \x03(\x0b\x32\x18.io.openaristos.ResourceB\x08\xfa\x42\x05\x92\x01\x02\x08\x01\x12\x31\n\x08metadata\x18\t \x01(\x0b\x32\x1f.io.openaristos.AdapterMetadata\"\x16\n\x14\x41pplyAdapterResponse\".\n\x15RetractAdapterRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x18\n\x16RetractAdapterResponse\"b\n\x12ListAdapterRequest\x12\r\n\x05query\x18\x01 \x01(\t\x12=\n\x12pagination_options\x18\x02 \x01(\x0b\x32!.io.openaristos.PaginationOptions\"\xd7\x02\n\x13ListAdapterResponse\x12\x44\n\x08\x61\x64\x61pters\x18\x01 \x03(\x0b\x32\x32.io.openaristos.ListAdapterResponse.AdapterDetails\x1a\xf9\x01\n\x0e\x41\x64\x61pterDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x04 \x01(\t\x12\x42\n\x11\x65xecution_profile\x18\x05 \x01(\x0b\x32\'.io.openaristos.AdapterExecutionProfile\x12\x35\n\nparameters\x18\x06 \x03(\x0b\x32!.io.openaristos.AdapterParameters\x12+\n\tresources\x18\x07 \x03(\x0b\x32\x18.io.openaristos.Resource\"-\n\x14\x45nableAdapterRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x17\n\x15\x45nableAdapterResponse\".\n\x15\x44isableAdapterRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x18\n\x16\x44isableAdapterResponse\"L\n\x14\x41pplyResourceRequest\x12\x34\n\x08resource\x18\x01 \x01(\x0b\x32\x18.io.openaristos.ResourceB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01\"\x17\n\x15\x41pplyResourceResponse\"/\n\x16RetractResourceRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x19\n\x17RetractResourceResponse\"c\n\x13ListResourceRequest\x12\r\n\x05query\x18\x01 \x01(\t\x12=\n\x12pagination_options\x18\x02 \x01(\x0b\x32!.io.openaristos.PaginationOptions\"C\n\x14ListResourceResponse\x12+\n\tresources\x18\x01 \x03(\x0b\x32\x18.io.openaristos.Resource\".\n\x15\x45nableResourceRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x18\n\x16\x45nableResourceResponse\"/\n\x16\x44isableResourceRequest\x12\x15\n\x04\x63ode\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\"\x19\n\x17\x44isableResourceResponse\"m\n\x11\x41\x64\x61pterParameters\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x65\x66\x61ult\x18\x02 \x01(\t\x12:\n\tdata_type\x18\x03 \x01(\x0e\x32\x1d.io.openaristos.AttributeTypeB\x08\xfa\x42\x05\x82\x01\x02\x10\x01\"v\n\x1c\x41\x64\x61pterExecutionContextEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12:\n\tdata_type\x18\x03 \x01(\x0e\x32\x1d.io.openaristos.AttributeTypeB\x08\xfa\x42\x05\x82\x01\x02\x10\x01\"\x93\x01\n\x17\x41\x64\x61pterExecutionProfile\x12\x1c\n\x0b\x65ngine_kind\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x11\n\tstreaming\x18\x02 \x01(\x08\x12G\n\x07\x63ontext\x18\x03 \x03(\x0b\x32,.io.openaristos.AdapterExecutionContextEntryB\x08\xfa\x42\x05\x92\x01\x02\x08\x01\"\x9f\x02\n\x1d\x43reateIngestionSessionRequest\x12\x18\n\x07\x61\x64\x61pter\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x1d\n\x0clanding_zone\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12!\n\x10mastering_engine\x18\x03 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x02\x12\x1c\n\nwarehouses\x18\x04 \x03(\tB\x08\xfa\x42\x05\x92\x01\x02\x08\x01\x12Q\n\nparameters\x18\x05 \x03(\x0b\x32=.io.openaristos.CreateIngestionSessionRequest.ParametersEntry\x1a\x31\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\",\n\x1e\x43reateIngestionSessionResponse\x12\n\n\x02id\x18\x01 \x01(\r\"\xb3\x01\n\x1e\x45xecuteIngestionSessionRequest\x12\n\n\x02id\x18\x01 \x01(\r\x12R\n\nparameters\x18\x05 \x03(\x0b\x32>.io.openaristos.ExecuteIngestionSessionRequest.ParametersEntry\x1a\x31\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"(\n\x1aGetIngestionSessionRequest\x12\n\n\x02id\x18\x01 \x01(\r\"\xc9\x01\n\x1eIngestionSessionStatusResponse\x12\n\n\x02id\x18\x01 \x01(\r\x12O\n\x06status\x18\x02 \x01(\x0e\x32\x35.io.openaristos.IngestionSessionStatusResponse.StatusB\x08\xfa\x42\x05\x82\x01\x02\x10\x01\"J\n\x06Status\x12\x07\n\x03NEW\x10\x00\x12\x0f\n\x0bIN_PROGRESS\x10\x01\x12\x0c\n\x08\x43OMPLETE\x10\x02\x12\t\n\x05\x45RROR\x10\x03\x12\r\n\tCANCELLED\x10\x04\"\x93\x01\n\x0cMasterEntity\x12\x0b\n\x03uid\x18\x01 \x01(\t\x12\x42\n\x0b\x64\x65scriptors\x18\x02 \x03(\x0b\x32-.io.openaristos.MasterEntity.DescriptorsEntry\x1a\x32\n\x10\x44\x65scriptorsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"=\n\x1dMasterIngestionSessionRequest\x12\x1c\n\x14ingestion_session_id\x18\x01 \x01(\r\" \n\x1eMasterIngestionSessionResponse\"\x82\x01\n&PublishIngestionSessionResourceRequest\x12\x1c\n\x14ingestion_session_id\x18\x01 \x01(\r\x12\x15\n\rresource_code\x18\x02 \x01(\t\x12\x16\n\x0ewarehouse_code\x18\x03 \x01(\t\x12\x0b\n\x03\x61ll\x18\x04 \x01(\x08\")\n\'PublishIngestionSessionResourceResponse\"{\n,PublishIngestionSessionResourceStatusRequest\x12\x1c\n\x14ingestion_session_id\x18\x01 \x01(\r\x12\x15\n\rresource_code\x18\x02 \x01(\t\x12\x16\n\x0ewarehouse_code\x18\x03 \x01(\t\"\x91\x02\n%PublishIngestionSessionStatusResponse\x12\x1c\n\x14ingestion_session_id\x18\x01 \x01(\r\x12\x15\n\rresource_code\x18\x02 \x01(\t\x12\x16\n\x0ewarehouse_code\x18\x03 \x01(\t\x12V\n\x06status\x18\x04 \x01(\x0e\x32<.io.openaristos.PublishIngestionSessionStatusResponse.StatusB\x08\xfa\x42\x05\x82\x01\x02\x10\x01\"C\n\x06Status\x12\x0f\n\x0bUNPUBLISHED\x10\x00\x12\x0e\n\nPUBLISHING\x10\x01\x12\r\n\tPUBLISHED\x10\x02\x12\t\n\x05\x45RROR\x10\x03\"w\n\x1fRetractPublishingSessionRequest\x12\x1c\n\x14ingestion_session_id\x18\x01 \x01(\r\x12\x11\n\tresources\x18\x02 \x03(\t\x12\x16\n\x0ewarehouse_code\x18\x05 \x01(\t\x12\x0b\n\x03\x61ll\x18\x06 \x01(\x08\"\"\n RetractPublishingSessionResponse\">\n\x1eRetractMasteringSessionRequest\x12\x1c\n\x14ingestion_session_id\x18\x01 \x01(\r\"!\n\x1fRetractMasteringSessionResponse\"1\n\x1fHaltIncrementalMasteringRequest\x12\x0e\n\x06\x65ngine\x18\x01 \x01(\t\"\"\n HaltIncrementalMasteringResponse\"3\n!ResumeIncrementalMasteringRequest\x12\x0e\n\x06\x65ngine\x18\x01 \x01(\t\"$\n\"ResumeIncrementalMasteringResponse\"\xb7\x01\n\x1aResolveResourcePathRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\twarehouse\x18\x02 \x01(\t\x12H\n\x07\x63ontext\x18\x03 \x03(\x0b\x32\x37.io.openaristos.ResolveResourcePathRequest.ContextEntry\x1a.\n\x0c\x43ontextEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xb0\x01\n\x1bResolveResourcePathResponse\x12\x16\n\x0ewarehouse_kind\x18\x01 \x01(\t\x12I\n\x07\x63ontext\x18\x02 \x03(\x0b\x32\x38.io.openaristos.ResolveResourcePathResponse.ContextEntry\x1a.\n\x0c\x43ontextEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x0b\n\tPSRequest\"\xa7\x02\n\nPSResponse\x12G\n\x12ingestion_sessions\x18\x01 \x03(\x0b\x32+.io.openaristos.PSResponse.IngestionSession\x12\x43\n\x10publish_sessions\x18\x02 \x03(\x0b\x32).io.openaristos.PSResponse.PublishSession\x1a.\n\x10IngestionSession\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0e\n\x06status\x18\x02 \x01(\t\x1a[\n\x0ePublishSession\x12\x14\n\x0cingestion_id\x18\x01 \x01(\r\x12\x11\n\twarehouse\x18\x02 \x01(\t\x12\x10\n\x08resource\x18\x03 \x01(\t\x12\x0e\n\x06status\x18\x04 \x01(\t\"D\n$ListIngestionSessionArtifactsRequest\x12\x1c\n\x14ingestion_session_id\x18\x01 \x01(\r\"\x94\x01\n%ListIngestionSessionArtifactsResponse\x12\x17\n\x0f\x61ssets_location\x18\x01 \x01(\t\x12\x1e\n\x16relationships_location\x18\x02 \x01(\t\x12\x19\n\x11\x65ntities_location\x18\x03 \x01(\t\x12\x17\n\x0f\x65rrors_location\x18\x04 \x01(\t\"t\n\x1cListIngestionSessionsRequest\x12)\n\x05since\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12)\n\x05until\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xab\x01\n\x1dListIngestionSessionsResponse\x12Z\n\x12ingestion_sessions\x18\x01 \x03(\x0b\x32>.io.openaristos.ListIngestionSessionsResponse.IngestionSession\x1a.\n\x10IngestionSession\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0e\n\x06status\x18\x02 \x01(\t\"U\n\x17\x41pplyPerspectiveRequest\x12:\n\x0bperspective\x18\x01 \x01(\x0b\x32\x1b.io.openaristos.PerspectiveB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01\"\x1a\n\x18\x41pplyPerspectiveResponse\"E\n\x19RetractPerspectiveRequest\x12\x0e\n\x06\x65ngine\x18\x01 \x01(\t\x12\x18\n\x10perspective_code\x18\x02 \x01(\t\"\x1c\n\x1aRetractPerspectiveResponse\"v\n\x16ListPerspectiveRequest\x12\x0e\n\x06\x65ngine\x18\x01 \x01(\t\x12\r\n\x05query\x18\x02 \x01(\t\x12=\n\x12pagination_options\x18\x03 \x01(\x0b\x32!.io.openaristos.PaginationOptions\"L\n\x17ListPerspectiveResponse\x12\x31\n\x0cperspectives\x18\x01 \x03(\x0b\x32\x1b.io.openaristos.Perspective\"N\n\rEventMetadata\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12/\n\x0boccurred_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x92\x01\n\x1aPublishResourceFailedEvent\x12/\n\x08metadata\x18\x01 \x01(\x0b\x32\x1d.io.openaristos.EventMetadata\x12\x10\n\x08resource\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65stination\x18\x03 \x01(\t\x12\x1c\n\x14ingestion_session_id\x18\x04 \x01(\r\"\x94\x01\n\x1cPublishResourceCompleteEvent\x12/\n\x08metadata\x18\x01 \x01(\x0b\x32\x1d.io.openaristos.EventMetadata\x12\x10\n\x08resource\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65stination\x18\x03 \x01(\t\x12\x1c\n\x14ingestion_session_id\x18\x04 \x01(\r\"\x98\x01\n$ExecuteIngestionSessionCompleteEvent\x12/\n\x08metadata\x18\x01 \x01(\x0b\x32\x1d.io.openaristos.EventMetadata\x12?\n\x07request\x18\x02 \x01(\x0b\x32..io.openaristos.ExecuteIngestionSessionRequest\"\x96\x01\n\"ExecuteIngestionSessionFailedEvent\x12/\n\x08metadata\x18\x01 \x01(\x0b\x32\x1d.io.openaristos.EventMetadata\x12?\n\x07request\x18\x02 \x01(\x0b\x32..io.openaristos.ExecuteIngestionSessionRequest\"\xad\x01\n\x1cIngestionSessionCreatedEvent\x12/\n\x08metadata\x18\x01 \x01(\x0b\x32\x1d.io.openaristos.EventMetadata\x12\x1c\n\x14ingestion_session_id\x18\x02 \x01(\r\x12>\n\x07request\x18\x03 \x01(\x0b\x32-.io.openaristos.CreateIngestionSessionRequest\"~\n\x16ResolveUniverseRequest\x12*\n\x08universe\x18\x01 \x01(\x0b\x32\x18.io.openaristos.Universe\x12\x38\n\x10\x65\x66\x66\x65\x63tive_dating\x18\x02 \x01(\x0b\x32\x1e.io.openaristos.TimestampRange\"R\n\x19ResolveUniverseGroupEntry\x12\x35\n\x07\x65ntries\x18\x01 \x03(\x0b\x32$.io.openaristos.ResolveUniverseEntry\"\xad\x02\n\x14ResolveUniverseEntry\x12\x11\n\tattribute\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12J\n\tdata_type\x18\x03 \x01(\x0e\x32-.io.openaristos.ResolveUniverseEntry.DataTypeB\x08\xfa\x42\x05\x82\x01\x02\x10\x01\"\xa6\x01\n\x08\x44\x61taType\x12\n\n\x06STRING\x10\x00\x12\x0b\n\x07INTEGER\x10\x01\x12\x0b\n\x07NUMERIC\x10\x02\x12\r\n\tTIMESTAMP\x10\x03\x12\x0b\n\x07\x42OOLEAN\x10\x04\x12\x0c\n\x08\x46LOATING\x10\x05\x12\t\n\x05\x42YTES\x10\x06\x12\x08\n\x04\x44\x41TE\x10\x07\x12\x0c\n\x08\x44\x41TETIME\x10\x08\x12\x08\n\x04TIME\x10\t\x12\x13\n\x0fTIMESTAMP_RANGE\x10\n\x12\x08\n\x04JSON\x10\x0b\"\xcc\x01\n\nErrorEntry\x12\x0f\n\x07message\x18\x01 \x01(\t\x12/\n\x0boccurred_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\nstacktrace\x18\x03 \x01(\t\x12\x38\n\x07\x63ontext\x18\x04 \x03(\x0b\x32\'.io.openaristos.ErrorEntry.ContextEntry\x1a.\n\x0c\x43ontextEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*B\n\x0e\x43omparisonType\x12\x06\n\x02\x45Q\x10\x00\x12\x06\n\x02LT\x10\x01\x12\x07\n\x03LTE\x10\x02\x12\x06\n\x02GT\x10\x03\x12\x07\n\x03GTE\x10\x04\x12\x06\n\x02NE\x10\x05*\xab\x01\n\rAttributeType\x12\n\n\x06STRING\x10\x00\x12\x0b\n\x07INTEGER\x10\x01\x12\x0b\n\x07NUMERIC\x10\x02\x12\r\n\tTIMESTAMP\x10\x03\x12\x0b\n\x07\x42OOLEAN\x10\x04\x12\x0c\n\x08\x46LOATING\x10\x05\x12\t\n\x05\x42YTES\x10\x06\x12\x08\n\x04\x44\x41TE\x10\x07\x12\x0c\n\x08\x44\x41TETIME\x10\x08\x12\x08\n\x04TIME\x10\t\x12\x13\n\x0fTIMESTAMP_RANGE\x10\n\x12\x08\n\x04JSON\x10\x0b\x32\x95,\n\x08\x42\x61kplane\x12Y\n\x0c\x41pplyAdapter\x12#.io.openaristos.ApplyAdapterRequest\x1a$.io.openaristos.ApplyAdapterResponse\x12_\n\x0eRetractAdapter\x12%.io.openaristos.RetractAdapterRequest\x1a&.io.openaristos.RetractAdapterResponse\x12V\n\x0bListAdapter\x12\".io.openaristos.ListAdapterRequest\x1a#.io.openaristos.ListAdapterResponse\x12\\\n\rEnableAdapter\x12$.io.openaristos.EnableAdapterRequest\x1a%.io.openaristos.EnableAdapterResponse\x12_\n\x0e\x44isableAdapter\x12%.io.openaristos.DisableAdapterRequest\x1a&.io.openaristos.DisableAdapterResponse\x12\\\n\rApplyResource\x12$.io.openaristos.ApplyResourceRequest\x1a%.io.openaristos.ApplyResourceResponse\x12\x62\n\x0fRetractResource\x12&.io.openaristos.RetractResourceRequest\x1a\'.io.openaristos.RetractResourceResponse\x12Z\n\rListResources\x12#.io.openaristos.ListResourceRequest\x1a$.io.openaristos.ListResourceResponse\x12_\n\x0e\x45nableResource\x12%.io.openaristos.EnableResourceRequest\x1a&.io.openaristos.EnableResourceResponse\x12\x62\n\x0f\x44isableResource\x12&.io.openaristos.DisableResourceRequest\x1a\'.io.openaristos.DisableResourceResponse\x12V\n\x0fMigrateResource\x12 .io.openaristos.MigrationRequest\x1a!.io.openaristos.MigrationResponse\x12\x64\n\x17ResourceMigrationStatus\x12&.io.openaristos.MigrationStatusRequest\x1a!.io.openaristos.MigrationResponse\x12\x64\n\x0fResolveResource\x12&.io.openaristos.ResolveResourceRequest\x1a\'.io.openaristos.ResolveResourceResponse0\x01\x12n\n\x15ResolveResourceIntent\x12&.io.openaristos.ResolveResourceRequest\x1a-.io.openaristos.ResolveResourceIntentResponse\x12\x66\n\x0fResolveUniverse\x12&.io.openaristos.ResolveUniverseRequest\x1a).io.openaristos.ResolveUniverseGroupEntry0\x01\x12\x65\n\x10\x41pplyLandingZone\x12\'.io.openaristos.ApplyLandingZoneRequest\x1a(.io.openaristos.ApplyLandingZoneResponse\x12k\n\x12RetractLandingZone\x12).io.openaristos.RetractLandingZoneRequest\x1a*.io.openaristos.RetractLandingZoneResponse\x12\x62\n\x0fListLandingZone\x12&.io.openaristos.ListLandingZoneRequest\x1a\'.io.openaristos.ListLandingZoneResponse\x12h\n\x11\x45nableLandingZone\x12(.io.openaristos.EnableLandingZoneRequest\x1a).io.openaristos.EnableLandingZoneResponse\x12k\n\x12\x44isableLandingZone\x12).io.openaristos.DisableLandingZoneRequest\x1a*.io.openaristos.DisableLandingZoneResponse\x12\\\n\rApplyOntology\x12$.io.openaristos.ApplyOntologyRequest\x1a%.io.openaristos.ApplyOntologyResponse\x12\x62\n\x0fRetractOntology\x12&.io.openaristos.RetractOntologyRequest\x1a\'.io.openaristos.RetractOntologyResponse\x12Y\n\x0cListOntology\x12#.io.openaristos.ListOntologyRequest\x1a$.io.openaristos.ListOntologyResponse\x12q\n\x14\x41pplyExecutionEngine\x12+.io.openaristos.ApplyExecutionEngineRequest\x1a,.io.openaristos.ApplyExecutionEngineResponse\x12w\n\x16RetractExecutionEngine\x12-.io.openaristos.RetractExecutionEngineRequest\x1a..io.openaristos.RetractExecutionEngineResponse\x12n\n\x13ListExecutionEngine\x12*.io.openaristos.ListExecutionEngineRequest\x1a+.io.openaristos.ListExecutionEngineResponse\x12t\n\x15\x45nableExecutionEngine\x12,.io.openaristos.EnableExecutionEngineRequest\x1a-.io.openaristos.EnableExecutionEngineResponse\x12w\n\x16\x44isableExecutionEngine\x12-.io.openaristos.DisableExecutionEngineRequest\x1a..io.openaristos.DisableExecutionEngineResponse\x12_\n\x0e\x41pplyWarehouse\x12%.io.openaristos.ApplyWarehouseRequest\x1a&.io.openaristos.ApplyWarehouseResponse\x12\x65\n\x10RetractWarehouse\x12\'.io.openaristos.RetractWarehouseRequest\x1a(.io.openaristos.RetractWarehouseResponse\x12\\\n\rListWarehouse\x12$.io.openaristos.ListWarehouseRequest\x1a%.io.openaristos.ListWarehouseResponse\x12\x62\n\x0f\x45nableWarehouse\x12&.io.openaristos.EnableWarehouseRequest\x1a\'.io.openaristos.EnableWarehouseResponse\x12\x65\n\x10\x44isableWarehouse\x12\'.io.openaristos.DisableWarehouseRequest\x1a(.io.openaristos.DisableWarehouseResponse\x12q\n\x14\x41pplyMasteringEngine\x12+.io.openaristos.ApplyMasteringEngineRequest\x1a,.io.openaristos.ApplyMasteringEngineResponse\x12w\n\x16RetractMasteringEngine\x12-.io.openaristos.RetractMasteringEngineRequest\x1a..io.openaristos.RetractMasteringEngineResponse\x12w\n\x16\x43reateIngestionSession\x12-.io.openaristos.CreateIngestionSessionRequest\x1a..io.openaristos.CreateIngestionSessionResponse\x12y\n\x17\x45xecuteIngestionSession\x12..io.openaristos.ExecuteIngestionSessionRequest\x1a..io.openaristos.IngestionSessionStatusResponse\x12q\n\x13GetIngestionSession\x12*.io.openaristos.GetIngestionSessionRequest\x1a..io.openaristos.IngestionSessionStatusResponse\x12w\n\x16MasterIngestionSession\x12-.io.openaristos.MasterIngestionSessionRequest\x1a..io.openaristos.MasterIngestionSessionResponse\x12}\n\x18HaltIncrementalMastering\x12/.io.openaristos.HaltIncrementalMasteringRequest\x1a\x30.io.openaristos.HaltIncrementalMasteringResponse\x12\x83\x01\n\x1aResumeIncrementalMastering\x12\x31.io.openaristos.ResumeIncrementalMasteringRequest\x1a\x32.io.openaristos.ResumeIncrementalMasteringResponse\x12\x92\x01\n\x1fPublishIngestionSessionResource\x12\x36.io.openaristos.PublishIngestionSessionResourceRequest\x1a\x37.io.openaristos.PublishIngestionSessionResourceResponse\x12\x9c\x01\n%PublishIngestionSessionResourceStatus\x12<.io.openaristos.PublishIngestionSessionResourceStatusRequest\x1a\x35.io.openaristos.PublishIngestionSessionStatusResponse\x12}\n\x18RetractPublishingSession\x12/.io.openaristos.RetractPublishingSessionRequest\x1a\x30.io.openaristos.RetractPublishingSessionResponse\x12z\n\x17RetractMasteringSession\x12..io.openaristos.RetractMasteringSessionRequest\x1a/.io.openaristos.RetractMasteringSessionResponse\x12n\n\x13ResolveResourcePath\x12*.io.openaristos.ResolveResourcePathRequest\x1a+.io.openaristos.ResolveResourcePathResponse\x12\x8c\x01\n\x1dListIngestionSessionArtifacts\x12\x34.io.openaristos.ListIngestionSessionArtifactsRequest\x1a\x35.io.openaristos.ListIngestionSessionArtifactsResponse\x12t\n\x15ListIngestionSessions\x12,.io.openaristos.ListIngestionSessionsRequest\x1a-.io.openaristos.ListIngestionSessionsResponse\x12\x42\n\tSessionPS\x12\x19.io.openaristos.PSRequest\x1a\x1a.io.openaristos.PSResponse\x12\x65\n\x10\x41pplyPerspective\x12\'.io.openaristos.ApplyPerspectiveRequest\x1a(.io.openaristos.ApplyPerspectiveResponse\x12k\n\x12RetractPerspective\x12).io.openaristos.RetractPerspectiveRequest\x1a*.io.openaristos.RetractPerspectiveResponse\x12\x62\n\x0fListPerspective\x12&.io.openaristos.ListPerspectiveRequest\x1a\'.io.openaristos.ListPerspectiveResponseB6\n\x0eio.openaristosB\rBakplaneProtoP\x01Z\x0eproto/bakplane\xa2\x02\x02\x42Pb\x06proto3'
  ,
  dependencies=[bakplane_dot_validate_dot_validate__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])

_COMPARISONTYPE = _descriptor.EnumDescriptor(
  name='ComparisonType',
  full_name='io.openaristos.ComparisonType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EQ', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LTE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GT', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GTE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NE', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15572,
  serialized_end=15638,
)
_sym_db.RegisterEnumDescriptor(_COMPARISONTYPE)

ComparisonType = enum_type_wrapper.EnumTypeWrapper(_COMPARISONTYPE)
_ATTRIBUTETYPE = _descriptor.EnumDescriptor(
  name='AttributeType',
  full_name='io.openaristos.AttributeType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STRING', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTEGER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NUMERIC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIMESTAMP', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FLOATING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BYTES', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATE', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATETIME', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIMESTAMP_RANGE', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JSON', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15641,
  serialized_end=15812,
)
_sym_db.RegisterEnumDescriptor(_ATTRIBUTETYPE)

AttributeType = enum_type_wrapper.EnumTypeWrapper(_ATTRIBUTETYPE)
EQ = 0
LT = 1
LTE = 2
GT = 3
GTE = 4
NE = 5
STRING = 0
INTEGER = 1
NUMERIC = 2
TIMESTAMP = 3
BOOLEAN = 4
FLOATING = 5
BYTES = 6
DATE = 7
DATETIME = 8
TIME = 9
TIMESTAMP_RANGE = 10
JSON = 11


_MIGRATIONREQUEST_DIRECTION = _descriptor.EnumDescriptor(
  name='Direction',
  full_name='io.openaristos.MigrationRequest.Direction',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UP', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DOWN', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=884,
  serialized_end=913,
)
_sym_db.RegisterEnumDescriptor(_MIGRATIONREQUEST_DIRECTION)

_SCHEMAATTRIBUTE_MAPPINGTYPE = _descriptor.EnumDescriptor(
  name='MappingType',
  full_name='io.openaristos.SchemaAttribute.MappingType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='IDENTIFIER', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PARTIAL_IDENTIFIER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DESC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IGNORED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=3574,
  serialized_end=3650,
)
_sym_db.RegisterEnumDescriptor(_SCHEMAATTRIBUTE_MAPPINGTYPE)

_INGESTIONSESSIONSTATUSRESPONSE_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='io.openaristos.IngestionSessionStatusResponse.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NEW', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IN_PROGRESS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPLETE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11048,
  serialized_end=11122,
)
_sym_db.RegisterEnumDescriptor(_INGESTIONSESSIONSTATUSRESPONSE_STATUS)

_PUBLISHINGESTIONSESSIONSTATUSRESPONSE_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='io.openaristos.PublishIngestionSessionStatusResponse.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNPUBLISHED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PUBLISHING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PUBLISHED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11879,
  serialized_end=11946,
)
_sym_db.RegisterEnumDescriptor(_PUBLISHINGESTIONSESSIONSTATUSRESPONSE_STATUS)

_RESOLVEUNIVERSEENTRY_DATATYPE = _descriptor.EnumDescriptor(
  name='DataType',
  full_name='io.openaristos.ResolveUniverseEntry.DataType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STRING', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTEGER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NUMERIC', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIMESTAMP', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FLOATING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BYTES', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATE', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATETIME', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIMESTAMP_RANGE', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='JSON', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15197,
  serialized_end=15363,
)
_sym_db.RegisterEnumDescriptor(_RESOLVEUNIVERSEENTRY_DATATYPE)


_TIMESTAMPRANGE = _descriptor.Descriptor(
  name='TimestampRange',
  full_name='io.openaristos.TimestampRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_dt', full_name='io.openaristos.TimestampRange.start_dt', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\262\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_dt', full_name='io.openaristos.TimestampRange.end_dt', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\262\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=110,
  serialized_end=236,
)


_PAGINATIONOPTIONS = _descriptor.Descriptor(
  name='PaginationOptions',
  full_name='io.openaristos.PaginationOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='per_page', full_name='io.openaristos.PaginationOptions.per_page', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page', full_name='io.openaristos.PaginationOptions.page', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=238,
  serialized_end=289,
)


_UNIVERSE = _descriptor.Descriptor(
  name='Universe',
  full_name='io.openaristos.Universe',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='definition', full_name='io.openaristos.Universe.definition', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dialect', full_name='io.openaristos.Universe.dialect', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='engine', full_name='io.openaristos.Universe.engine', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=291,
  serialized_end=363,
)


_PERSPECTIVE = _descriptor.Descriptor(
  name='Perspective',
  full_name='io.openaristos.Perspective',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.Perspective.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.Perspective.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.Perspective.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='weight', full_name='io.openaristos.Perspective.weight', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\006*\004\030d(\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='permanent_attributes', full_name='io.openaristos.Perspective.permanent_attributes', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=366,
  serialized_end=550,
)


_PERSPECTIVEPERMANENTATTRIBUTE = _descriptor.Descriptor(
  name='PerspectivePermanentAttribute',
  full_name='io.openaristos.PerspectivePermanentAttribute',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entity_type_code', full_name='io.openaristos.PerspectivePermanentAttribute.entity_type_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attribute_code', full_name='io.openaristos.PerspectivePermanentAttribute.attribute_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=552,
  serialized_end=651,
)


_MIGRATIONREQUEST_FORCE = _descriptor.Descriptor(
  name='Force',
  full_name='io.openaristos.MigrationRequest.Force',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='io.openaristos.MigrationRequest.Force.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=858,
  serialized_end=882,
)

_MIGRATIONREQUEST = _descriptor.Descriptor(
  name='MigrationRequest',
  full_name='io.openaristos.MigrationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.MigrationRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.MigrationRequest.warehouse_kind', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='direction', full_name='io.openaristos.MigrationRequest.direction', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\202\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='force', full_name='io.openaristos.MigrationRequest.force', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MIGRATIONREQUEST_FORCE, ],
  enum_types=[
    _MIGRATIONREQUEST_DIRECTION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=654,
  serialized_end=913,
)


_MIGRATIONSTATUSREQUEST = _descriptor.Descriptor(
  name='MigrationStatusRequest',
  full_name='io.openaristos.MigrationStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.MigrationStatusRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.MigrationStatusRequest.warehouse_kind', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=915,
  serialized_end=995,
)


_MIGRATIONRESPONSE = _descriptor.Descriptor(
  name='MigrationResponse',
  full_name='io.openaristos.MigrationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='io.openaristos.MigrationResponse.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.MigrationResponse.warehouse_kind', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.MigrationResponse.code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=997,
  serialized_end=1071,
)


_RESOLVERESOURCEREQUEST_POINTERSENTRY = _descriptor.Descriptor(
  name='PointersEntry',
  full_name='io.openaristos.ResolveResourceRequest.PointersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ResolveResourceRequest.PointersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ResolveResourceRequest.PointersEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1317,
  serialized_end=1364,
)

_RESOLVERESOURCEREQUEST = _descriptor.Descriptor(
  name='ResolveResourceRequest',
  full_name='io.openaristos.ResolveResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='universe', full_name='io.openaristos.ResolveResourceRequest.universe', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\212\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='io.openaristos.ResolveResourceRequest.path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse', full_name='io.openaristos.ResolveResourceRequest.warehouse', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='effective_dating', full_name='io.openaristos.ResolveResourceRequest.effective_dating', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pointers', full_name='io.openaristos.ResolveResourceRequest.pointers', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_RESOLVERESOURCEREQUEST_POINTERSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1074,
  serialized_end=1364,
)


_RESOLVERESOURCEMAPPINGCONTEXTENTRY = _descriptor.Descriptor(
  name='ResolveResourceMappingContextEntry',
  full_name='io.openaristos.ResolveResourceMappingContextEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_field_name', full_name='io.openaristos.ResolveResourceMappingContextEntry.source_field_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_field_expression', full_name='io.openaristos.ResolveResourceMappingContextEntry.target_field_expression', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attribute_type', full_name='io.openaristos.ResolveResourceMappingContextEntry.attribute_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='used_for_mapping', full_name='io.openaristos.ResolveResourceMappingContextEntry.used_for_mapping', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='map_to_field_name', full_name='io.openaristos.ResolveResourceMappingContextEntry.map_to_field_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_effective_dating', full_name='io.openaristos.ResolveResourceMappingContextEntry.is_effective_dating', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='comparison_type', full_name='io.openaristos.ResolveResourceMappingContextEntry.comparison_type', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1367,
  serialized_end=1657,
)


_RESOLVERESOURCERESPONSE = _descriptor.Descriptor(
  name='ResolveResourceResponse',
  full_name='io.openaristos.ResolveResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='row', full_name='io.openaristos.ResolveResourceResponse.row', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1659,
  serialized_end=1697,
)


_RESOLVERESOURCEINTENTRESPONSE_CONTEXTENTRY = _descriptor.Descriptor(
  name='ContextEntry',
  full_name='io.openaristos.ResolveResourceIntentResponse.ContextEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ResolveResourceIntentResponse.ContextEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ResolveResourceIntentResponse.ContextEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2042,
  serialized_end=2088,
)

_RESOLVERESOURCEINTENTRESPONSE = _descriptor.Descriptor(
  name='ResolveResourceIntentResponse',
  full_name='io.openaristos.ResolveResourceIntentResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.ResolveResourceIntentResponse.warehouse_kind', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table_name', full_name='io.openaristos.ResolveResourceIntentResponse.table_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_name', full_name='io.openaristos.ResolveResourceIntentResponse.schema_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mapping', full_name='io.openaristos.ResolveResourceIntentResponse.mapping', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='columns', full_name='io.openaristos.ResolveResourceIntentResponse.columns', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='io.openaristos.ResolveResourceIntentResponse.context', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='effective_dating', full_name='io.openaristos.ResolveResourceIntentResponse.effective_dating', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_RESOLVERESOURCEINTENTRESPONSE_CONTEXTENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1700,
  serialized_end=2088,
)


_RESOLVERESOURCEINTENTMAPPING = _descriptor.Descriptor(
  name='ResolveResourceIntentMapping',
  full_name='io.openaristos.ResolveResourceIntentMapping',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='table_name', full_name='io.openaristos.ResolveResourceIntentMapping.table_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_name', full_name='io.openaristos.ResolveResourceIntentMapping.schema_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='columns', full_name='io.openaristos.ResolveResourceIntentMapping.columns', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clauses', full_name='io.openaristos.ResolveResourceIntentMapping.clauses', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2091,
  serialized_end=2277,
)


_COLUMNENTRY = _descriptor.Descriptor(
  name='ColumnEntry',
  full_name='io.openaristos.ColumnEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.ColumnEntry.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataType', full_name='io.openaristos.ColumnEntry.dataType', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_effective_dating', full_name='io.openaristos.ColumnEntry.is_effective_dating', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2279,
  serialized_end=2353,
)


_RESOLVERESOURCEINTENTMAPPINGCLAUSE = _descriptor.Descriptor(
  name='ResolveResourceIntentMappingClause',
  full_name='io.openaristos.ResolveResourceIntentMappingClause',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource_field_name', full_name='io.openaristos.ResolveResourceIntentMappingClause.resource_field_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mapping_field_name', full_name='io.openaristos.ResolveResourceIntentMappingClause.mapping_field_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='comparison', full_name='io.openaristos.ResolveResourceIntentMappingClause.comparison', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2356,
  serialized_end=2501,
)


_CONFIGURATIONENTRY = _descriptor.Descriptor(
  name='ConfigurationEntry',
  full_name='io.openaristos.ConfigurationEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ConfigurationEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ConfigurationEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2503,
  serialized_end=2569,
)


_APPLYLANDINGZONEREQUEST = _descriptor.Descriptor(
  name='ApplyLandingZoneRequest',
  full_name='io.openaristos.ApplyLandingZoneRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ApplyLandingZoneRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plugin', full_name='io.openaristos.ApplyLandingZoneRequest.plugin', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configurations', full_name='io.openaristos.ApplyLandingZoneRequest.configurations', index=2,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2571,
  serialized_end=2695,
)


_APPLYLANDINGZONERESPONSE = _descriptor.Descriptor(
  name='ApplyLandingZoneResponse',
  full_name='io.openaristos.ApplyLandingZoneResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2697,
  serialized_end=2723,
)


_RETRACTLANDINGZONEREQUEST = _descriptor.Descriptor(
  name='RetractLandingZoneRequest',
  full_name='io.openaristos.RetractLandingZoneRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RetractLandingZoneRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2725,
  serialized_end=2775,
)


_RETRACTLANDINGZONERESPONSE = _descriptor.Descriptor(
  name='RetractLandingZoneResponse',
  full_name='io.openaristos.RetractLandingZoneResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2777,
  serialized_end=2805,
)


_LISTLANDINGZONEREQUEST = _descriptor.Descriptor(
  name='ListLandingZoneRequest',
  full_name='io.openaristos.ListLandingZoneRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='io.openaristos.ListLandingZoneRequest.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_options', full_name='io.openaristos.ListLandingZoneRequest.pagination_options', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2807,
  serialized_end=2918,
)


_LISTLANDINGZONERESPONSE_LANDINGZONEDETAILS = _descriptor.Descriptor(
  name='LandingZoneDetails',
  full_name='io.openaristos.ListLandingZoneResponse.LandingZoneDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ListLandingZoneResponse.LandingZoneDetails.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plugin', full_name='io.openaristos.ListLandingZoneResponse.LandingZoneDetails.plugin', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='active', full_name='io.openaristos.ListLandingZoneResponse.LandingZoneDetails.active', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configurations', full_name='io.openaristos.ListLandingZoneResponse.LandingZoneDetails.configurations', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3031,
  serialized_end=3157,
)

_LISTLANDINGZONERESPONSE = _descriptor.Descriptor(
  name='ListLandingZoneResponse',
  full_name='io.openaristos.ListLandingZoneResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='landing_zones', full_name='io.openaristos.ListLandingZoneResponse.landing_zones', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_LISTLANDINGZONERESPONSE_LANDINGZONEDETAILS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2921,
  serialized_end=3157,
)


_ENABLELANDINGZONEREQUEST = _descriptor.Descriptor(
  name='EnableLandingZoneRequest',
  full_name='io.openaristos.EnableLandingZoneRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.EnableLandingZoneRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3159,
  serialized_end=3208,
)


_ENABLELANDINGZONERESPONSE = _descriptor.Descriptor(
  name='EnableLandingZoneResponse',
  full_name='io.openaristos.EnableLandingZoneResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3210,
  serialized_end=3237,
)


_DISABLELANDINGZONEREQUEST = _descriptor.Descriptor(
  name='DisableLandingZoneRequest',
  full_name='io.openaristos.DisableLandingZoneRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.DisableLandingZoneRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3239,
  serialized_end=3289,
)


_DISABLELANDINGZONERESPONSE = _descriptor.Descriptor(
  name='DisableLandingZoneResponse',
  full_name='io.openaristos.DisableLandingZoneResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3291,
  serialized_end=3319,
)


_SCHEMAATTRIBUTE = _descriptor.Descriptor(
  name='SchemaAttribute',
  full_name='io.openaristos.SchemaAttribute',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.SchemaAttribute.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.SchemaAttribute.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.SchemaAttribute.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='io.openaristos.SchemaAttribute.data_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mode', full_name='io.openaristos.SchemaAttribute.mode', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B r\036R\010nullableR\010requiredR\010repeated', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='indexable', full_name='io.openaristos.SchemaAttribute.indexable', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mapping_type', full_name='io.openaristos.SchemaAttribute.mapping_type', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\202\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SCHEMAATTRIBUTE_MAPPINGTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3322,
  serialized_end=3650,
)


_SCHEMA = _descriptor.Descriptor(
  name='Schema',
  full_name='io.openaristos.Schema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.Schema.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.Schema.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.Schema.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attributes', full_name='io.openaristos.Schema.attributes', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\222\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3653,
  serialized_end=3782,
)


_ENTITYTYPE = _descriptor.Descriptor(
  name='EntityType',
  full_name='io.openaristos.EntityType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.EntityType.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.EntityType.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.EntityType.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='io.openaristos.EntityType.schema', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\212\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3784,
  serialized_end=3904,
)


_RELATIONSHIPTYPE = _descriptor.Descriptor(
  name='RelationshipType',
  full_name='io.openaristos.RelationshipType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.RelationshipType.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RelationshipType.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.RelationshipType.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_entity_type', full_name='io.openaristos.RelationshipType.source_entity_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_entity_type', full_name='io.openaristos.RelationshipType.target_entity_type', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3907,
  serialized_end=4057,
)


_APPLYONTOLOGYREQUEST = _descriptor.Descriptor(
  name='ApplyOntologyRequest',
  full_name='io.openaristos.ApplyOntologyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.ApplyOntologyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ApplyOntologyRequest.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.ApplyOntologyRequest.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entities', full_name='io.openaristos.ApplyOntologyRequest.entities', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\222\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='relationships', full_name='io.openaristos.ApplyOntologyRequest.relationships', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4060,
  serialized_end=4253,
)


_APPLYONTOLOGYRESPONSE = _descriptor.Descriptor(
  name='ApplyOntologyResponse',
  full_name='io.openaristos.ApplyOntologyResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4255,
  serialized_end=4278,
)


_RETRACTONTOLOGYREQUEST = _descriptor.Descriptor(
  name='RetractOntologyRequest',
  full_name='io.openaristos.RetractOntologyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='mastering_engine_code', full_name='io.openaristos.RetractOntologyRequest.mastering_engine_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RetractOntologyRequest.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='force', full_name='io.openaristos.RetractOntologyRequest.force', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4280,
  serialized_end=4382,
)


_RETRACTONTOLOGYRESPONSE = _descriptor.Descriptor(
  name='RetractOntologyResponse',
  full_name='io.openaristos.RetractOntologyResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4384,
  serialized_end=4409,
)


_LISTONTOLOGYREQUEST = _descriptor.Descriptor(
  name='ListOntologyRequest',
  full_name='io.openaristos.ListOntologyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='io.openaristos.ListOntologyRequest.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_options', full_name='io.openaristos.ListOntologyRequest.pagination_options', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4411,
  serialized_end=4519,
)


_LISTONTOLOGYRESPONSE = _descriptor.Descriptor(
  name='ListOntologyResponse',
  full_name='io.openaristos.ListOntologyResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entities', full_name='io.openaristos.ListOntologyResponse.entities', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='relationships', full_name='io.openaristos.ListOntologyResponse.relationships', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4521,
  serialized_end=4646,
)


_EXECUTIONENGINE = _descriptor.Descriptor(
  name='ExecutionEngine',
  full_name='io.openaristos.ExecutionEngine',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.ExecutionEngine.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ExecutionEngine.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.ExecutionEngine.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4648,
  serialized_end=4723,
)


_APPLYEXECUTIONENGINEREQUEST = _descriptor.Descriptor(
  name='ApplyExecutionEngineRequest',
  full_name='io.openaristos.ApplyExecutionEngineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ApplyExecutionEngineRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='kind', full_name='io.openaristos.ApplyExecutionEngineRequest.kind', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plugin', full_name='io.openaristos.ApplyExecutionEngineRequest.plugin', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configurations', full_name='io.openaristos.ApplyExecutionEngineRequest.configurations', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4726,
  serialized_end=4868,
)


_APPLYEXECUTIONENGINERESPONSE = _descriptor.Descriptor(
  name='ApplyExecutionEngineResponse',
  full_name='io.openaristos.ApplyExecutionEngineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4870,
  serialized_end=4900,
)


_RETRACTEXECUTIONENGINEREQUEST = _descriptor.Descriptor(
  name='RetractExecutionEngineRequest',
  full_name='io.openaristos.RetractExecutionEngineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RetractExecutionEngineRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4902,
  serialized_end=4956,
)


_RETRACTEXECUTIONENGINERESPONSE = _descriptor.Descriptor(
  name='RetractExecutionEngineResponse',
  full_name='io.openaristos.RetractExecutionEngineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4958,
  serialized_end=4990,
)


_LISTEXECUTIONENGINEREQUEST = _descriptor.Descriptor(
  name='ListExecutionEngineRequest',
  full_name='io.openaristos.ListExecutionEngineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='io.openaristos.ListExecutionEngineRequest.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_options', full_name='io.openaristos.ListExecutionEngineRequest.pagination_options', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4992,
  serialized_end=5098,
)


_LISTEXECUTIONENGINERESPONSE = _descriptor.Descriptor(
  name='ListExecutionEngineResponse',
  full_name='io.openaristos.ListExecutionEngineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='engines', full_name='io.openaristos.ListExecutionEngineResponse.engines', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5100,
  serialized_end=5179,
)


_ENABLEEXECUTIONENGINEREQUEST = _descriptor.Descriptor(
  name='EnableExecutionEngineRequest',
  full_name='io.openaristos.EnableExecutionEngineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.EnableExecutionEngineRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5181,
  serialized_end=5234,
)


_ENABLEEXECUTIONENGINERESPONSE = _descriptor.Descriptor(
  name='EnableExecutionEngineResponse',
  full_name='io.openaristos.EnableExecutionEngineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5236,
  serialized_end=5267,
)


_DISABLEEXECUTIONENGINEREQUEST = _descriptor.Descriptor(
  name='DisableExecutionEngineRequest',
  full_name='io.openaristos.DisableExecutionEngineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.DisableExecutionEngineRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5269,
  serialized_end=5323,
)


_DISABLEEXECUTIONENGINERESPONSE = _descriptor.Descriptor(
  name='DisableExecutionEngineResponse',
  full_name='io.openaristos.DisableExecutionEngineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5325,
  serialized_end=5357,
)


_WAREHOUSE = _descriptor.Descriptor(
  name='Warehouse',
  full_name='io.openaristos.Warehouse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.Warehouse.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.Warehouse.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.Warehouse.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5359,
  serialized_end=5428,
)


_APPLYWAREHOUSEREQUEST = _descriptor.Descriptor(
  name='ApplyWarehouseRequest',
  full_name='io.openaristos.ApplyWarehouseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ApplyWarehouseRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plugin', full_name='io.openaristos.ApplyWarehouseRequest.plugin', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='kind', full_name='io.openaristos.ApplyWarehouseRequest.kind', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configurations', full_name='io.openaristos.ApplyWarehouseRequest.configurations', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5431,
  serialized_end=5567,
)


_APPLYWAREHOUSERESPONSE = _descriptor.Descriptor(
  name='ApplyWarehouseResponse',
  full_name='io.openaristos.ApplyWarehouseResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5569,
  serialized_end=5593,
)


_RETRACTWAREHOUSEREQUEST = _descriptor.Descriptor(
  name='RetractWarehouseRequest',
  full_name='io.openaristos.RetractWarehouseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RetractWarehouseRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5595,
  serialized_end=5643,
)


_RETRACTWAREHOUSERESPONSE = _descriptor.Descriptor(
  name='RetractWarehouseResponse',
  full_name='io.openaristos.RetractWarehouseResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5645,
  serialized_end=5671,
)


_LISTWAREHOUSEREQUEST = _descriptor.Descriptor(
  name='ListWarehouseRequest',
  full_name='io.openaristos.ListWarehouseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='io.openaristos.ListWarehouseRequest.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_options', full_name='io.openaristos.ListWarehouseRequest.pagination_options', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5673,
  serialized_end=5773,
)


_LISTWAREHOUSERESPONSE = _descriptor.Descriptor(
  name='ListWarehouseResponse',
  full_name='io.openaristos.ListWarehouseResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='warehouses', full_name='io.openaristos.ListWarehouseResponse.warehouses', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5775,
  serialized_end=5845,
)


_ENABLEWAREHOUSEREQUEST = _descriptor.Descriptor(
  name='EnableWarehouseRequest',
  full_name='io.openaristos.EnableWarehouseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.EnableWarehouseRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5847,
  serialized_end=5894,
)


_ENABLEWAREHOUSERESPONSE = _descriptor.Descriptor(
  name='EnableWarehouseResponse',
  full_name='io.openaristos.EnableWarehouseResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5896,
  serialized_end=5921,
)


_DISABLEWAREHOUSEREQUEST = _descriptor.Descriptor(
  name='DisableWarehouseRequest',
  full_name='io.openaristos.DisableWarehouseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.DisableWarehouseRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5923,
  serialized_end=5971,
)


_DISABLEWAREHOUSERESPONSE = _descriptor.Descriptor(
  name='DisableWarehouseResponse',
  full_name='io.openaristos.DisableWarehouseResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5973,
  serialized_end=5999,
)


_APPLYMASTERINGENGINEREQUEST = _descriptor.Descriptor(
  name='ApplyMasteringEngineRequest',
  full_name='io.openaristos.ApplyMasteringEngineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ApplyMasteringEngineRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plugin', full_name='io.openaristos.ApplyMasteringEngineRequest.plugin', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='configurations', full_name='io.openaristos.ApplyMasteringEngineRequest.configurations', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6002,
  serialized_end=6130,
)


_APPLYMASTERINGENGINERESPONSE = _descriptor.Descriptor(
  name='ApplyMasteringEngineResponse',
  full_name='io.openaristos.ApplyMasteringEngineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6132,
  serialized_end=6162,
)


_RETRACTMASTERINGENGINEREQUEST = _descriptor.Descriptor(
  name='RetractMasteringEngineRequest',
  full_name='io.openaristos.RetractMasteringEngineRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RetractMasteringEngineRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6164,
  serialized_end=6218,
)


_RETRACTMASTERINGENGINERESPONSE = _descriptor.Descriptor(
  name='RetractMasteringEngineResponse',
  full_name='io.openaristos.RetractMasteringEngineResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6220,
  serialized_end=6252,
)


_RESOURCESTORECONTEXTENTRY = _descriptor.Descriptor(
  name='ResourceStoreContextEntry',
  full_name='io.openaristos.ResourceStoreContextEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ResourceStoreContextEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ResourceStoreContextEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6254,
  serialized_end=6309,
)


_RESOURCESTORE = _descriptor.Descriptor(
  name='ResourceStore',
  full_name='io.openaristos.ResourceStore',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ResourceStore.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='virtual', full_name='io.openaristos.ResourceStore.virtual', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse', full_name='io.openaristos.ResourceStore.warehouse', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='io.openaristos.ResourceStore.context', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6312,
  serialized_end=6446,
)


_RESOURCEPOINTERENTRY = _descriptor.Descriptor(
  name='ResourcePointerEntry',
  full_name='io.openaristos.ResourcePointerEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ResourcePointerEntry.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity_type', full_name='io.openaristos.ResourcePointerEntry.entity_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6448,
  serialized_end=6523,
)


_RESOURCEMETADATAENTRY = _descriptor.Descriptor(
  name='ResourceMetadataEntry',
  full_name='io.openaristos.ResourceMetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ResourceMetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ResourceMetadataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6525,
  serialized_end=6576,
)


_RESOURCECONTEXTPOINTERENTRY = _descriptor.Descriptor(
  name='ResourceContextPointerEntry',
  full_name='io.openaristos.ResourceContextPointerEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ResourceContextPointerEntry.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='column_name', full_name='io.openaristos.ResourceContextPointerEntry.column_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6578,
  serialized_end=6642,
)


_RESOURCEDATING = _descriptor.Descriptor(
  name='ResourceDating',
  full_name='io.openaristos.ResourceDating',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='effective_start_dt', full_name='io.openaristos.ResourceDating.effective_start_dt', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='effective_end_dt', full_name='io.openaristos.ResourceDating.effective_end_dt', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='knowledge_start_dt', full_name='io.openaristos.ResourceDating.knowledge_start_dt', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='knowledge_end_dt', full_name='io.openaristos.ResourceDating.knowledge_end_dt', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6644,
  serialized_end=6768,
)


_RESOURCEHASHING = _descriptor.Descriptor(
  name='ResourceHashing',
  full_name='io.openaristos.ResourceHashing',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='knowledge_hash', full_name='io.openaristos.ResourceHashing.knowledge_hash', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload_hash', full_name='io.openaristos.ResourceHashing.payload_hash', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6770,
  serialized_end=6833,
)


_RESOURCEPUBLISHING = _descriptor.Descriptor(
  name='ResourcePublishing',
  full_name='io.openaristos.ResourcePublishing',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ddl_create_staging_table', full_name='io.openaristos.ResourcePublishing.ddl_create_staging_table', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ddl_create_staging_table_indexes', full_name='io.openaristos.ResourcePublishing.ddl_create_staging_table_indexes', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ddl_drop_staging_table', full_name='io.openaristos.ResourcePublishing.ddl_drop_staging_table', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dml_obsolete_knowledge_query', full_name='io.openaristos.ResourcePublishing.dml_obsolete_knowledge_query', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dml_incremental_query', full_name='io.openaristos.ResourcePublishing.dml_incremental_query', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6836,
  serialized_end=7033,
)


_RESOURCERETRACTION = _descriptor.Descriptor(
  name='ResourceRetraction',
  full_name='io.openaristos.ResourceRetraction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.ResourceRetraction.warehouse_kind', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='steps', full_name='io.openaristos.ResourceRetraction.steps', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7035,
  serialized_end=7094,
)


_RESOURCECONTEXT = _descriptor.Descriptor(
  name='ResourceContext',
  full_name='io.openaristos.ResourceContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.ResourceContext.warehouse_kind', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pointers', full_name='io.openaristos.ResourceContext.pointers', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dating', full_name='io.openaristos.ResourceContext.dating', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hashing', full_name='io.openaristos.ResourceContext.hashing', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='io.openaristos.ResourceContext.metadata', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='publishing', full_name='io.openaristos.ResourceContext.publishing', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='io.openaristos.ResourceContext.dataset', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retraction', full_name='io.openaristos.ResourceContext.retraction', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7097,
  serialized_end=7485,
)


_RESOURCE = _descriptor.Descriptor(
  name='Resource',
  full_name='io.openaristos.Resource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.Resource.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='kind', full_name='io.openaristos.Resource.kind', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perspective_code', full_name='io.openaristos.Resource.perspective_code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.Resource.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.Resource.name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='io.openaristos.Resource.dataset', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='io.openaristos.Resource.schema', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\212\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pointers', full_name='io.openaristos.Resource.pointers', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stores', full_name='io.openaristos.Resource.stores', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retention_hours', full_name='io.openaristos.Resource.retention_hours', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='io.openaristos.Resource.context', index=10,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='migrations', full_name='io.openaristos.Resource.migrations', index=11,
      number=12, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\222\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_virtual', full_name='io.openaristos.Resource.is_virtual', index=12,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7488,
  serialized_end=7927,
)


_ADAPTER = _descriptor.Descriptor(
  name='Adapter',
  full_name='io.openaristos.Adapter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.Adapter.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.Adapter.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.Adapter.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='author', full_name='io.openaristos.Adapter.author', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7929,
  serialized_end=8012,
)


_ADAPTERPERSISTENCE = _descriptor.Descriptor(
  name='AdapterPersistence',
  full_name='io.openaristos.AdapterPersistence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.AdapterPersistence.warehouse_kind', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entries', full_name='io.openaristos.AdapterPersistence.entries', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8014,
  serialized_end=8116,
)


_ADAPTERPERSISTENCEENTRY = _descriptor.Descriptor(
  name='AdapterPersistenceEntry',
  full_name='io.openaristos.AdapterPersistenceEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.AdapterPersistenceEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.AdapterPersistenceEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8118,
  serialized_end=8171,
)


_ADAPTERMETADATA = _descriptor.Descriptor(
  name='AdapterMetadata',
  full_name='io.openaristos.AdapterMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='persistence', full_name='io.openaristos.AdapterMetadata.persistence', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retraction', full_name='io.openaristos.AdapterMetadata.retraction', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8174,
  serialized_end=8304,
)


_MIGRATION = _descriptor.Descriptor(
  name='Migration',
  full_name='io.openaristos.Migration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.Migration.warehouse_kind', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='migrations_uri', full_name='io.openaristos.Migration.migrations_uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_name', full_name='io.openaristos.Migration.schema_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table_name', full_name='io.openaristos.Migration.table_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8306,
  serialized_end=8406,
)


_APPLYADAPTERREQUEST = _descriptor.Descriptor(
  name='ApplyAdapterRequest',
  full_name='io.openaristos.ApplyAdapterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.ApplyAdapterRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\003', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ApplyAdapterRequest.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.ApplyAdapterRequest.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='author', full_name='io.openaristos.ApplyAdapterRequest.author', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='provider', full_name='io.openaristos.ApplyAdapterRequest.provider', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='execution_profile', full_name='io.openaristos.ApplyAdapterRequest.execution_profile', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\212\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='io.openaristos.ApplyAdapterRequest.parameters', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resources', full_name='io.openaristos.ApplyAdapterRequest.resources', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\222\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='io.openaristos.ApplyAdapterRequest.metadata', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8409,
  serialized_end=8788,
)


_APPLYADAPTERRESPONSE = _descriptor.Descriptor(
  name='ApplyAdapterResponse',
  full_name='io.openaristos.ApplyAdapterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8790,
  serialized_end=8812,
)


_RETRACTADAPTERREQUEST = _descriptor.Descriptor(
  name='RetractAdapterRequest',
  full_name='io.openaristos.RetractAdapterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RetractAdapterRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8814,
  serialized_end=8860,
)


_RETRACTADAPTERRESPONSE = _descriptor.Descriptor(
  name='RetractAdapterResponse',
  full_name='io.openaristos.RetractAdapterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8862,
  serialized_end=8886,
)


_LISTADAPTERREQUEST = _descriptor.Descriptor(
  name='ListAdapterRequest',
  full_name='io.openaristos.ListAdapterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='io.openaristos.ListAdapterRequest.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_options', full_name='io.openaristos.ListAdapterRequest.pagination_options', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8888,
  serialized_end=8986,
)


_LISTADAPTERRESPONSE_ADAPTERDETAILS = _descriptor.Descriptor(
  name='AdapterDetails',
  full_name='io.openaristos.ListAdapterResponse.AdapterDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='io.openaristos.ListAdapterResponse.AdapterDetails.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.ListAdapterResponse.AdapterDetails.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='io.openaristos.ListAdapterResponse.AdapterDetails.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='author', full_name='io.openaristos.ListAdapterResponse.AdapterDetails.author', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='execution_profile', full_name='io.openaristos.ListAdapterResponse.AdapterDetails.execution_profile', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='io.openaristos.ListAdapterResponse.AdapterDetails.parameters', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resources', full_name='io.openaristos.ListAdapterResponse.AdapterDetails.resources', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9083,
  serialized_end=9332,
)

_LISTADAPTERRESPONSE = _descriptor.Descriptor(
  name='ListAdapterResponse',
  full_name='io.openaristos.ListAdapterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='adapters', full_name='io.openaristos.ListAdapterResponse.adapters', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_LISTADAPTERRESPONSE_ADAPTERDETAILS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8989,
  serialized_end=9332,
)


_ENABLEADAPTERREQUEST = _descriptor.Descriptor(
  name='EnableAdapterRequest',
  full_name='io.openaristos.EnableAdapterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.EnableAdapterRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9334,
  serialized_end=9379,
)


_ENABLEADAPTERRESPONSE = _descriptor.Descriptor(
  name='EnableAdapterResponse',
  full_name='io.openaristos.EnableAdapterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9381,
  serialized_end=9404,
)


_DISABLEADAPTERREQUEST = _descriptor.Descriptor(
  name='DisableAdapterRequest',
  full_name='io.openaristos.DisableAdapterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.DisableAdapterRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9406,
  serialized_end=9452,
)


_DISABLEADAPTERRESPONSE = _descriptor.Descriptor(
  name='DisableAdapterResponse',
  full_name='io.openaristos.DisableAdapterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9454,
  serialized_end=9478,
)


_APPLYRESOURCEREQUEST = _descriptor.Descriptor(
  name='ApplyResourceRequest',
  full_name='io.openaristos.ApplyResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='io.openaristos.ApplyResourceRequest.resource', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\212\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9480,
  serialized_end=9556,
)


_APPLYRESOURCERESPONSE = _descriptor.Descriptor(
  name='ApplyResourceResponse',
  full_name='io.openaristos.ApplyResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9558,
  serialized_end=9581,
)


_RETRACTRESOURCEREQUEST = _descriptor.Descriptor(
  name='RetractResourceRequest',
  full_name='io.openaristos.RetractResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.RetractResourceRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9583,
  serialized_end=9630,
)


_RETRACTRESOURCERESPONSE = _descriptor.Descriptor(
  name='RetractResourceResponse',
  full_name='io.openaristos.RetractResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9632,
  serialized_end=9657,
)


_LISTRESOURCEREQUEST = _descriptor.Descriptor(
  name='ListResourceRequest',
  full_name='io.openaristos.ListResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='io.openaristos.ListResourceRequest.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_options', full_name='io.openaristos.ListResourceRequest.pagination_options', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9659,
  serialized_end=9758,
)


_LISTRESOURCERESPONSE = _descriptor.Descriptor(
  name='ListResourceResponse',
  full_name='io.openaristos.ListResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resources', full_name='io.openaristos.ListResourceResponse.resources', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9760,
  serialized_end=9827,
)


_ENABLERESOURCEREQUEST = _descriptor.Descriptor(
  name='EnableResourceRequest',
  full_name='io.openaristos.EnableResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.EnableResourceRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9829,
  serialized_end=9875,
)


_ENABLERESOURCERESPONSE = _descriptor.Descriptor(
  name='EnableResourceResponse',
  full_name='io.openaristos.EnableResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9877,
  serialized_end=9901,
)


_DISABLERESOURCEREQUEST = _descriptor.Descriptor(
  name='DisableResourceRequest',
  full_name='io.openaristos.DisableResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='io.openaristos.DisableResourceRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9903,
  serialized_end=9950,
)


_DISABLERESOURCERESPONSE = _descriptor.Descriptor(
  name='DisableResourceResponse',
  full_name='io.openaristos.DisableResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9952,
  serialized_end=9977,
)


_ADAPTERPARAMETERS = _descriptor.Descriptor(
  name='AdapterParameters',
  full_name='io.openaristos.AdapterParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.AdapterParameters.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='default', full_name='io.openaristos.AdapterParameters.default', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='io.openaristos.AdapterParameters.data_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\202\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9979,
  serialized_end=10088,
)


_ADAPTEREXECUTIONCONTEXTENTRY = _descriptor.Descriptor(
  name='AdapterExecutionContextEntry',
  full_name='io.openaristos.AdapterExecutionContextEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.AdapterExecutionContextEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.AdapterExecutionContextEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='io.openaristos.AdapterExecutionContextEntry.data_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\202\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10090,
  serialized_end=10208,
)


_ADAPTEREXECUTIONPROFILE = _descriptor.Descriptor(
  name='AdapterExecutionProfile',
  full_name='io.openaristos.AdapterExecutionProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine_kind', full_name='io.openaristos.AdapterExecutionProfile.engine_kind', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='streaming', full_name='io.openaristos.AdapterExecutionProfile.streaming', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='io.openaristos.AdapterExecutionProfile.context', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\222\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10211,
  serialized_end=10358,
)


_CREATEINGESTIONSESSIONREQUEST_PARAMETERSENTRY = _descriptor.Descriptor(
  name='ParametersEntry',
  full_name='io.openaristos.CreateIngestionSessionRequest.ParametersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.CreateIngestionSessionRequest.ParametersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.CreateIngestionSessionRequest.ParametersEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10599,
  serialized_end=10648,
)

_CREATEINGESTIONSESSIONREQUEST = _descriptor.Descriptor(
  name='CreateIngestionSessionRequest',
  full_name='io.openaristos.CreateIngestionSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='adapter', full_name='io.openaristos.CreateIngestionSessionRequest.adapter', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='landing_zone', full_name='io.openaristos.CreateIngestionSessionRequest.landing_zone', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mastering_engine', full_name='io.openaristos.CreateIngestionSessionRequest.mastering_engine', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\004r\002\020\002', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouses', full_name='io.openaristos.CreateIngestionSessionRequest.warehouses', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\222\001\002\010\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='io.openaristos.CreateIngestionSessionRequest.parameters', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CREATEINGESTIONSESSIONREQUEST_PARAMETERSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10361,
  serialized_end=10648,
)


_CREATEINGESTIONSESSIONRESPONSE = _descriptor.Descriptor(
  name='CreateIngestionSessionResponse',
  full_name='io.openaristos.CreateIngestionSessionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='io.openaristos.CreateIngestionSessionResponse.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10650,
  serialized_end=10694,
)


_EXECUTEINGESTIONSESSIONREQUEST_PARAMETERSENTRY = _descriptor.Descriptor(
  name='ParametersEntry',
  full_name='io.openaristos.ExecuteIngestionSessionRequest.ParametersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ExecuteIngestionSessionRequest.ParametersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ExecuteIngestionSessionRequest.ParametersEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10599,
  serialized_end=10648,
)

_EXECUTEINGESTIONSESSIONREQUEST = _descriptor.Descriptor(
  name='ExecuteIngestionSessionRequest',
  full_name='io.openaristos.ExecuteIngestionSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='io.openaristos.ExecuteIngestionSessionRequest.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='io.openaristos.ExecuteIngestionSessionRequest.parameters', index=1,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EXECUTEINGESTIONSESSIONREQUEST_PARAMETERSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10697,
  serialized_end=10876,
)


_GETINGESTIONSESSIONREQUEST = _descriptor.Descriptor(
  name='GetIngestionSessionRequest',
  full_name='io.openaristos.GetIngestionSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='io.openaristos.GetIngestionSessionRequest.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10878,
  serialized_end=10918,
)


_INGESTIONSESSIONSTATUSRESPONSE = _descriptor.Descriptor(
  name='IngestionSessionStatusResponse',
  full_name='io.openaristos.IngestionSessionStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='io.openaristos.IngestionSessionStatusResponse.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='io.openaristos.IngestionSessionStatusResponse.status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\202\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _INGESTIONSESSIONSTATUSRESPONSE_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10921,
  serialized_end=11122,
)


_MASTERENTITY_DESCRIPTORSENTRY = _descriptor.Descriptor(
  name='DescriptorsEntry',
  full_name='io.openaristos.MasterEntity.DescriptorsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.MasterEntity.DescriptorsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.MasterEntity.DescriptorsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11222,
  serialized_end=11272,
)

_MASTERENTITY = _descriptor.Descriptor(
  name='MasterEntity',
  full_name='io.openaristos.MasterEntity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uid', full_name='io.openaristos.MasterEntity.uid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='descriptors', full_name='io.openaristos.MasterEntity.descriptors', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MASTERENTITY_DESCRIPTORSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11125,
  serialized_end=11272,
)


_MASTERINGESTIONSESSIONREQUEST = _descriptor.Descriptor(
  name='MasterIngestionSessionRequest',
  full_name='io.openaristos.MasterIngestionSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.MasterIngestionSessionRequest.ingestion_session_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11274,
  serialized_end=11335,
)


_MASTERINGESTIONSESSIONRESPONSE = _descriptor.Descriptor(
  name='MasterIngestionSessionResponse',
  full_name='io.openaristos.MasterIngestionSessionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11337,
  serialized_end=11369,
)


_PUBLISHINGESTIONSESSIONRESOURCEREQUEST = _descriptor.Descriptor(
  name='PublishIngestionSessionResourceRequest',
  full_name='io.openaristos.PublishIngestionSessionResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.PublishIngestionSessionResourceRequest.ingestion_session_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_code', full_name='io.openaristos.PublishIngestionSessionResourceRequest.resource_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse_code', full_name='io.openaristos.PublishIngestionSessionResourceRequest.warehouse_code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all', full_name='io.openaristos.PublishIngestionSessionResourceRequest.all', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11372,
  serialized_end=11502,
)


_PUBLISHINGESTIONSESSIONRESOURCERESPONSE = _descriptor.Descriptor(
  name='PublishIngestionSessionResourceResponse',
  full_name='io.openaristos.PublishIngestionSessionResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11504,
  serialized_end=11545,
)


_PUBLISHINGESTIONSESSIONRESOURCESTATUSREQUEST = _descriptor.Descriptor(
  name='PublishIngestionSessionResourceStatusRequest',
  full_name='io.openaristos.PublishIngestionSessionResourceStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.PublishIngestionSessionResourceStatusRequest.ingestion_session_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_code', full_name='io.openaristos.PublishIngestionSessionResourceStatusRequest.resource_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse_code', full_name='io.openaristos.PublishIngestionSessionResourceStatusRequest.warehouse_code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11547,
  serialized_end=11670,
)


_PUBLISHINGESTIONSESSIONSTATUSRESPONSE = _descriptor.Descriptor(
  name='PublishIngestionSessionStatusResponse',
  full_name='io.openaristos.PublishIngestionSessionStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.PublishIngestionSessionStatusResponse.ingestion_session_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_code', full_name='io.openaristos.PublishIngestionSessionStatusResponse.resource_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse_code', full_name='io.openaristos.PublishIngestionSessionStatusResponse.warehouse_code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='io.openaristos.PublishIngestionSessionStatusResponse.status', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\202\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PUBLISHINGESTIONSESSIONSTATUSRESPONSE_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11673,
  serialized_end=11946,
)


_RETRACTPUBLISHINGSESSIONREQUEST = _descriptor.Descriptor(
  name='RetractPublishingSessionRequest',
  full_name='io.openaristos.RetractPublishingSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.RetractPublishingSessionRequest.ingestion_session_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resources', full_name='io.openaristos.RetractPublishingSessionRequest.resources', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse_code', full_name='io.openaristos.RetractPublishingSessionRequest.warehouse_code', index=2,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all', full_name='io.openaristos.RetractPublishingSessionRequest.all', index=3,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11948,
  serialized_end=12067,
)


_RETRACTPUBLISHINGSESSIONRESPONSE = _descriptor.Descriptor(
  name='RetractPublishingSessionResponse',
  full_name='io.openaristos.RetractPublishingSessionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12069,
  serialized_end=12103,
)


_RETRACTMASTERINGSESSIONREQUEST = _descriptor.Descriptor(
  name='RetractMasteringSessionRequest',
  full_name='io.openaristos.RetractMasteringSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.RetractMasteringSessionRequest.ingestion_session_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12105,
  serialized_end=12167,
)


_RETRACTMASTERINGSESSIONRESPONSE = _descriptor.Descriptor(
  name='RetractMasteringSessionResponse',
  full_name='io.openaristos.RetractMasteringSessionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12169,
  serialized_end=12202,
)


_HALTINCREMENTALMASTERINGREQUEST = _descriptor.Descriptor(
  name='HaltIncrementalMasteringRequest',
  full_name='io.openaristos.HaltIncrementalMasteringRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine', full_name='io.openaristos.HaltIncrementalMasteringRequest.engine', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12204,
  serialized_end=12253,
)


_HALTINCREMENTALMASTERINGRESPONSE = _descriptor.Descriptor(
  name='HaltIncrementalMasteringResponse',
  full_name='io.openaristos.HaltIncrementalMasteringResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12255,
  serialized_end=12289,
)


_RESUMEINCREMENTALMASTERINGREQUEST = _descriptor.Descriptor(
  name='ResumeIncrementalMasteringRequest',
  full_name='io.openaristos.ResumeIncrementalMasteringRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine', full_name='io.openaristos.ResumeIncrementalMasteringRequest.engine', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12291,
  serialized_end=12342,
)


_RESUMEINCREMENTALMASTERINGRESPONSE = _descriptor.Descriptor(
  name='ResumeIncrementalMasteringResponse',
  full_name='io.openaristos.ResumeIncrementalMasteringResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12344,
  serialized_end=12380,
)


_RESOLVERESOURCEPATHREQUEST_CONTEXTENTRY = _descriptor.Descriptor(
  name='ContextEntry',
  full_name='io.openaristos.ResolveResourcePathRequest.ContextEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ResolveResourcePathRequest.ContextEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ResolveResourcePathRequest.ContextEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2042,
  serialized_end=2088,
)

_RESOLVERESOURCEPATHREQUEST = _descriptor.Descriptor(
  name='ResolveResourcePathRequest',
  full_name='io.openaristos.ResolveResourcePathRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='io.openaristos.ResolveResourcePathRequest.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse', full_name='io.openaristos.ResolveResourcePathRequest.warehouse', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='io.openaristos.ResolveResourcePathRequest.context', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_RESOLVERESOURCEPATHREQUEST_CONTEXTENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12383,
  serialized_end=12566,
)


_RESOLVERESOURCEPATHRESPONSE_CONTEXTENTRY = _descriptor.Descriptor(
  name='ContextEntry',
  full_name='io.openaristos.ResolveResourcePathResponse.ContextEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ResolveResourcePathResponse.ContextEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ResolveResourcePathResponse.ContextEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2042,
  serialized_end=2088,
)

_RESOLVERESOURCEPATHRESPONSE = _descriptor.Descriptor(
  name='ResolveResourcePathResponse',
  full_name='io.openaristos.ResolveResourcePathResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='warehouse_kind', full_name='io.openaristos.ResolveResourcePathResponse.warehouse_kind', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='io.openaristos.ResolveResourcePathResponse.context', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_RESOLVERESOURCEPATHRESPONSE_CONTEXTENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12569,
  serialized_end=12745,
)


_PSREQUEST = _descriptor.Descriptor(
  name='PSRequest',
  full_name='io.openaristos.PSRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12747,
  serialized_end=12758,
)


_PSRESPONSE_INGESTIONSESSION = _descriptor.Descriptor(
  name='IngestionSession',
  full_name='io.openaristos.PSResponse.IngestionSession',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='io.openaristos.PSResponse.IngestionSession.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='io.openaristos.PSResponse.IngestionSession.status', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12917,
  serialized_end=12963,
)

_PSRESPONSE_PUBLISHSESSION = _descriptor.Descriptor(
  name='PublishSession',
  full_name='io.openaristos.PSResponse.PublishSession',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_id', full_name='io.openaristos.PSResponse.PublishSession.ingestion_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse', full_name='io.openaristos.PSResponse.PublishSession.warehouse', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='io.openaristos.PSResponse.PublishSession.resource', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='io.openaristos.PSResponse.PublishSession.status', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12965,
  serialized_end=13056,
)

_PSRESPONSE = _descriptor.Descriptor(
  name='PSResponse',
  full_name='io.openaristos.PSResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_sessions', full_name='io.openaristos.PSResponse.ingestion_sessions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='publish_sessions', full_name='io.openaristos.PSResponse.publish_sessions', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PSRESPONSE_INGESTIONSESSION, _PSRESPONSE_PUBLISHSESSION, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12761,
  serialized_end=13056,
)


_LISTINGESTIONSESSIONARTIFACTSREQUEST = _descriptor.Descriptor(
  name='ListIngestionSessionArtifactsRequest',
  full_name='io.openaristos.ListIngestionSessionArtifactsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.ListIngestionSessionArtifactsRequest.ingestion_session_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13058,
  serialized_end=13126,
)


_LISTINGESTIONSESSIONARTIFACTSRESPONSE = _descriptor.Descriptor(
  name='ListIngestionSessionArtifactsResponse',
  full_name='io.openaristos.ListIngestionSessionArtifactsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='assets_location', full_name='io.openaristos.ListIngestionSessionArtifactsResponse.assets_location', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='relationships_location', full_name='io.openaristos.ListIngestionSessionArtifactsResponse.relationships_location', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entities_location', full_name='io.openaristos.ListIngestionSessionArtifactsResponse.entities_location', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errors_location', full_name='io.openaristos.ListIngestionSessionArtifactsResponse.errors_location', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13129,
  serialized_end=13277,
)


_LISTINGESTIONSESSIONSREQUEST = _descriptor.Descriptor(
  name='ListIngestionSessionsRequest',
  full_name='io.openaristos.ListIngestionSessionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='since', full_name='io.openaristos.ListIngestionSessionsRequest.since', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='until', full_name='io.openaristos.ListIngestionSessionsRequest.until', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13279,
  serialized_end=13395,
)


_LISTINGESTIONSESSIONSRESPONSE_INGESTIONSESSION = _descriptor.Descriptor(
  name='IngestionSession',
  full_name='io.openaristos.ListIngestionSessionsResponse.IngestionSession',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='io.openaristos.ListIngestionSessionsResponse.IngestionSession.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='io.openaristos.ListIngestionSessionsResponse.IngestionSession.status', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12917,
  serialized_end=12963,
)

_LISTINGESTIONSESSIONSRESPONSE = _descriptor.Descriptor(
  name='ListIngestionSessionsResponse',
  full_name='io.openaristos.ListIngestionSessionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ingestion_sessions', full_name='io.openaristos.ListIngestionSessionsResponse.ingestion_sessions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_LISTINGESTIONSESSIONSRESPONSE_INGESTIONSESSION, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13398,
  serialized_end=13569,
)


_APPLYPERSPECTIVEREQUEST = _descriptor.Descriptor(
  name='ApplyPerspectiveRequest',
  full_name='io.openaristos.ApplyPerspectiveRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='perspective', full_name='io.openaristos.ApplyPerspectiveRequest.perspective', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\212\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13571,
  serialized_end=13656,
)


_APPLYPERSPECTIVERESPONSE = _descriptor.Descriptor(
  name='ApplyPerspectiveResponse',
  full_name='io.openaristos.ApplyPerspectiveResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13658,
  serialized_end=13684,
)


_RETRACTPERSPECTIVEREQUEST = _descriptor.Descriptor(
  name='RetractPerspectiveRequest',
  full_name='io.openaristos.RetractPerspectiveRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine', full_name='io.openaristos.RetractPerspectiveRequest.engine', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perspective_code', full_name='io.openaristos.RetractPerspectiveRequest.perspective_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13686,
  serialized_end=13755,
)


_RETRACTPERSPECTIVERESPONSE = _descriptor.Descriptor(
  name='RetractPerspectiveResponse',
  full_name='io.openaristos.RetractPerspectiveResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13757,
  serialized_end=13785,
)


_LISTPERSPECTIVEREQUEST = _descriptor.Descriptor(
  name='ListPerspectiveRequest',
  full_name='io.openaristos.ListPerspectiveRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine', full_name='io.openaristos.ListPerspectiveRequest.engine', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='io.openaristos.ListPerspectiveRequest.query', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pagination_options', full_name='io.openaristos.ListPerspectiveRequest.pagination_options', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13787,
  serialized_end=13905,
)


_LISTPERSPECTIVERESPONSE = _descriptor.Descriptor(
  name='ListPerspectiveResponse',
  full_name='io.openaristos.ListPerspectiveResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='perspectives', full_name='io.openaristos.ListPerspectiveResponse.perspectives', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13907,
  serialized_end=13983,
)


_EVENTMETADATA = _descriptor.Descriptor(
  name='EventMetadata',
  full_name='io.openaristos.EventMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='io.openaristos.EventMetadata.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='occurred_at', full_name='io.openaristos.EventMetadata.occurred_at', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13985,
  serialized_end=14063,
)


_PUBLISHRESOURCEFAILEDEVENT = _descriptor.Descriptor(
  name='PublishResourceFailedEvent',
  full_name='io.openaristos.PublishResourceFailedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='io.openaristos.PublishResourceFailedEvent.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='io.openaristos.PublishResourceFailedEvent.resource', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='destination', full_name='io.openaristos.PublishResourceFailedEvent.destination', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.PublishResourceFailedEvent.ingestion_session_id', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14066,
  serialized_end=14212,
)


_PUBLISHRESOURCECOMPLETEEVENT = _descriptor.Descriptor(
  name='PublishResourceCompleteEvent',
  full_name='io.openaristos.PublishResourceCompleteEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='io.openaristos.PublishResourceCompleteEvent.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='io.openaristos.PublishResourceCompleteEvent.resource', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='destination', full_name='io.openaristos.PublishResourceCompleteEvent.destination', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.PublishResourceCompleteEvent.ingestion_session_id', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14215,
  serialized_end=14363,
)


_EXECUTEINGESTIONSESSIONCOMPLETEEVENT = _descriptor.Descriptor(
  name='ExecuteIngestionSessionCompleteEvent',
  full_name='io.openaristos.ExecuteIngestionSessionCompleteEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='io.openaristos.ExecuteIngestionSessionCompleteEvent.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request', full_name='io.openaristos.ExecuteIngestionSessionCompleteEvent.request', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14366,
  serialized_end=14518,
)


_EXECUTEINGESTIONSESSIONFAILEDEVENT = _descriptor.Descriptor(
  name='ExecuteIngestionSessionFailedEvent',
  full_name='io.openaristos.ExecuteIngestionSessionFailedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='io.openaristos.ExecuteIngestionSessionFailedEvent.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request', full_name='io.openaristos.ExecuteIngestionSessionFailedEvent.request', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14521,
  serialized_end=14671,
)


_INGESTIONSESSIONCREATEDEVENT = _descriptor.Descriptor(
  name='IngestionSessionCreatedEvent',
  full_name='io.openaristos.IngestionSessionCreatedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='io.openaristos.IngestionSessionCreatedEvent.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ingestion_session_id', full_name='io.openaristos.IngestionSessionCreatedEvent.ingestion_session_id', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request', full_name='io.openaristos.IngestionSessionCreatedEvent.request', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14674,
  serialized_end=14847,
)


_RESOLVEUNIVERSEREQUEST = _descriptor.Descriptor(
  name='ResolveUniverseRequest',
  full_name='io.openaristos.ResolveUniverseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='universe', full_name='io.openaristos.ResolveUniverseRequest.universe', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='effective_dating', full_name='io.openaristos.ResolveUniverseRequest.effective_dating', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14849,
  serialized_end=14975,
)


_RESOLVEUNIVERSEGROUPENTRY = _descriptor.Descriptor(
  name='ResolveUniverseGroupEntry',
  full_name='io.openaristos.ResolveUniverseGroupEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entries', full_name='io.openaristos.ResolveUniverseGroupEntry.entries', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14977,
  serialized_end=15059,
)


_RESOLVEUNIVERSEENTRY = _descriptor.Descriptor(
  name='ResolveUniverseEntry',
  full_name='io.openaristos.ResolveUniverseEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='attribute', full_name='io.openaristos.ResolveUniverseEntry.attribute', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ResolveUniverseEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='io.openaristos.ResolveUniverseEntry.data_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\372B\005\202\001\002\020\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RESOLVEUNIVERSEENTRY_DATATYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15062,
  serialized_end=15363,
)


_ERRORENTRY_CONTEXTENTRY = _descriptor.Descriptor(
  name='ContextEntry',
  full_name='io.openaristos.ErrorEntry.ContextEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='io.openaristos.ErrorEntry.ContextEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='io.openaristos.ErrorEntry.ContextEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2042,
  serialized_end=2088,
)

_ERRORENTRY = _descriptor.Descriptor(
  name='ErrorEntry',
  full_name='io.openaristos.ErrorEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='io.openaristos.ErrorEntry.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='occurred_at', full_name='io.openaristos.ErrorEntry.occurred_at', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stacktrace', full_name='io.openaristos.ErrorEntry.stacktrace', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='context', full_name='io.openaristos.ErrorEntry.context', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ERRORENTRY_CONTEXTENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15366,
  serialized_end=15570,
)

_TIMESTAMPRANGE.fields_by_name['start_dt'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPRANGE.fields_by_name['end_dt'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PERSPECTIVE.fields_by_name['permanent_attributes'].message_type = _PERSPECTIVEPERMANENTATTRIBUTE
_MIGRATIONREQUEST_FORCE.containing_type = _MIGRATIONREQUEST
_MIGRATIONREQUEST.fields_by_name['direction'].enum_type = _MIGRATIONREQUEST_DIRECTION
_MIGRATIONREQUEST.fields_by_name['force'].message_type = _MIGRATIONREQUEST_FORCE
_MIGRATIONREQUEST_DIRECTION.containing_type = _MIGRATIONREQUEST
_RESOLVERESOURCEREQUEST_POINTERSENTRY.containing_type = _RESOLVERESOURCEREQUEST
_RESOLVERESOURCEREQUEST.fields_by_name['universe'].message_type = _UNIVERSE
_RESOLVERESOURCEREQUEST.fields_by_name['effective_dating'].message_type = _TIMESTAMPRANGE
_RESOLVERESOURCEREQUEST.fields_by_name['pointers'].message_type = _RESOLVERESOURCEREQUEST_POINTERSENTRY
_RESOLVERESOURCEMAPPINGCONTEXTENTRY.fields_by_name['attribute_type'].enum_type = _ATTRIBUTETYPE
_RESOLVERESOURCEMAPPINGCONTEXTENTRY.fields_by_name['comparison_type'].enum_type = _COMPARISONTYPE
_RESOLVERESOURCEINTENTRESPONSE_CONTEXTENTRY.containing_type = _RESOLVERESOURCEINTENTRESPONSE
_RESOLVERESOURCEINTENTRESPONSE.fields_by_name['mapping'].message_type = _RESOLVERESOURCEINTENTMAPPING
_RESOLVERESOURCEINTENTRESPONSE.fields_by_name['columns'].message_type = _COLUMNENTRY
_RESOLVERESOURCEINTENTRESPONSE.fields_by_name['context'].message_type = _RESOLVERESOURCEINTENTRESPONSE_CONTEXTENTRY
_RESOLVERESOURCEINTENTRESPONSE.fields_by_name['effective_dating'].message_type = _TIMESTAMPRANGE
_RESOLVERESOURCEINTENTMAPPING.fields_by_name['columns'].message_type = _COLUMNENTRY
_RESOLVERESOURCEINTENTMAPPING.fields_by_name['clauses'].message_type = _RESOLVERESOURCEINTENTMAPPINGCLAUSE
_RESOLVERESOURCEINTENTMAPPINGCLAUSE.fields_by_name['comparison'].enum_type = _COMPARISONTYPE
_APPLYLANDINGZONEREQUEST.fields_by_name['configurations'].message_type = _CONFIGURATIONENTRY
_LISTLANDINGZONEREQUEST.fields_by_name['pagination_options'].message_type = _PAGINATIONOPTIONS
_LISTLANDINGZONERESPONSE_LANDINGZONEDETAILS.fields_by_name['configurations'].message_type = _CONFIGURATIONENTRY
_LISTLANDINGZONERESPONSE_LANDINGZONEDETAILS.containing_type = _LISTLANDINGZONERESPONSE
_LISTLANDINGZONERESPONSE.fields_by_name['landing_zones'].message_type = _LISTLANDINGZONERESPONSE_LANDINGZONEDETAILS
_SCHEMAATTRIBUTE.fields_by_name['mapping_type'].enum_type = _SCHEMAATTRIBUTE_MAPPINGTYPE
_SCHEMAATTRIBUTE_MAPPINGTYPE.containing_type = _SCHEMAATTRIBUTE
_SCHEMA.fields_by_name['attributes'].message_type = _SCHEMAATTRIBUTE
_ENTITYTYPE.fields_by_name['schema'].message_type = _SCHEMA
_APPLYONTOLOGYREQUEST.fields_by_name['entities'].message_type = _ENTITYTYPE
_APPLYONTOLOGYREQUEST.fields_by_name['relationships'].message_type = _RELATIONSHIPTYPE
_LISTONTOLOGYREQUEST.fields_by_name['pagination_options'].message_type = _PAGINATIONOPTIONS
_LISTONTOLOGYRESPONSE.fields_by_name['entities'].message_type = _ENTITYTYPE
_LISTONTOLOGYRESPONSE.fields_by_name['relationships'].message_type = _RELATIONSHIPTYPE
_APPLYEXECUTIONENGINEREQUEST.fields_by_name['configurations'].message_type = _CONFIGURATIONENTRY
_LISTEXECUTIONENGINEREQUEST.fields_by_name['pagination_options'].message_type = _PAGINATIONOPTIONS
_LISTEXECUTIONENGINERESPONSE.fields_by_name['engines'].message_type = _EXECUTIONENGINE
_APPLYWAREHOUSEREQUEST.fields_by_name['configurations'].message_type = _CONFIGURATIONENTRY
_LISTWAREHOUSEREQUEST.fields_by_name['pagination_options'].message_type = _PAGINATIONOPTIONS
_LISTWAREHOUSERESPONSE.fields_by_name['warehouses'].message_type = _WAREHOUSE
_APPLYMASTERINGENGINEREQUEST.fields_by_name['configurations'].message_type = _CONFIGURATIONENTRY
_RESOURCESTORE.fields_by_name['context'].message_type = _RESOURCESTORECONTEXTENTRY
_RESOURCECONTEXT.fields_by_name['pointers'].message_type = _RESOURCECONTEXTPOINTERENTRY
_RESOURCECONTEXT.fields_by_name['dating'].message_type = _RESOURCEDATING
_RESOURCECONTEXT.fields_by_name['hashing'].message_type = _RESOURCEHASHING
_RESOURCECONTEXT.fields_by_name['metadata'].message_type = _RESOURCEMETADATAENTRY
_RESOURCECONTEXT.fields_by_name['publishing'].message_type = _RESOURCEPUBLISHING
_RESOURCECONTEXT.fields_by_name['retraction'].message_type = _RESOURCERETRACTION
_RESOURCE.fields_by_name['schema'].message_type = _SCHEMA
_RESOURCE.fields_by_name['pointers'].message_type = _RESOURCEPOINTERENTRY
_RESOURCE.fields_by_name['stores'].message_type = _RESOURCESTORE
_RESOURCE.fields_by_name['context'].message_type = _RESOURCECONTEXT
_RESOURCE.fields_by_name['migrations'].message_type = _MIGRATION
_ADAPTERPERSISTENCE.fields_by_name['entries'].message_type = _ADAPTERPERSISTENCEENTRY
_ADAPTERMETADATA.fields_by_name['persistence'].message_type = _ADAPTERPERSISTENCE
_ADAPTERMETADATA.fields_by_name['retraction'].message_type = _RESOURCERETRACTION
_APPLYADAPTERREQUEST.fields_by_name['execution_profile'].message_type = _ADAPTEREXECUTIONPROFILE
_APPLYADAPTERREQUEST.fields_by_name['parameters'].message_type = _ADAPTERPARAMETERS
_APPLYADAPTERREQUEST.fields_by_name['resources'].message_type = _RESOURCE
_APPLYADAPTERREQUEST.fields_by_name['metadata'].message_type = _ADAPTERMETADATA
_LISTADAPTERREQUEST.fields_by_name['pagination_options'].message_type = _PAGINATIONOPTIONS
_LISTADAPTERRESPONSE_ADAPTERDETAILS.fields_by_name['execution_profile'].message_type = _ADAPTEREXECUTIONPROFILE
_LISTADAPTERRESPONSE_ADAPTERDETAILS.fields_by_name['parameters'].message_type = _ADAPTERPARAMETERS
_LISTADAPTERRESPONSE_ADAPTERDETAILS.fields_by_name['resources'].message_type = _RESOURCE
_LISTADAPTERRESPONSE_ADAPTERDETAILS.containing_type = _LISTADAPTERRESPONSE
_LISTADAPTERRESPONSE.fields_by_name['adapters'].message_type = _LISTADAPTERRESPONSE_ADAPTERDETAILS
_APPLYRESOURCEREQUEST.fields_by_name['resource'].message_type = _RESOURCE
_LISTRESOURCEREQUEST.fields_by_name['pagination_options'].message_type = _PAGINATIONOPTIONS
_LISTRESOURCERESPONSE.fields_by_name['resources'].message_type = _RESOURCE
_ADAPTERPARAMETERS.fields_by_name['data_type'].enum_type = _ATTRIBUTETYPE
_ADAPTEREXECUTIONCONTEXTENTRY.fields_by_name['data_type'].enum_type = _ATTRIBUTETYPE
_ADAPTEREXECUTIONPROFILE.fields_by_name['context'].message_type = _ADAPTEREXECUTIONCONTEXTENTRY
_CREATEINGESTIONSESSIONREQUEST_PARAMETERSENTRY.containing_type = _CREATEINGESTIONSESSIONREQUEST
_CREATEINGESTIONSESSIONREQUEST.fields_by_name['parameters'].message_type = _CREATEINGESTIONSESSIONREQUEST_PARAMETERSENTRY
_EXECUTEINGESTIONSESSIONREQUEST_PARAMETERSENTRY.containing_type = _EXECUTEINGESTIONSESSIONREQUEST
_EXECUTEINGESTIONSESSIONREQUEST.fields_by_name['parameters'].message_type = _EXECUTEINGESTIONSESSIONREQUEST_PARAMETERSENTRY
_INGESTIONSESSIONSTATUSRESPONSE.fields_by_name['status'].enum_type = _INGESTIONSESSIONSTATUSRESPONSE_STATUS
_INGESTIONSESSIONSTATUSRESPONSE_STATUS.containing_type = _INGESTIONSESSIONSTATUSRESPONSE
_MASTERENTITY_DESCRIPTORSENTRY.containing_type = _MASTERENTITY
_MASTERENTITY.fields_by_name['descriptors'].message_type = _MASTERENTITY_DESCRIPTORSENTRY
_PUBLISHINGESTIONSESSIONSTATUSRESPONSE.fields_by_name['status'].enum_type = _PUBLISHINGESTIONSESSIONSTATUSRESPONSE_STATUS
_PUBLISHINGESTIONSESSIONSTATUSRESPONSE_STATUS.containing_type = _PUBLISHINGESTIONSESSIONSTATUSRESPONSE
_RESOLVERESOURCEPATHREQUEST_CONTEXTENTRY.containing_type = _RESOLVERESOURCEPATHREQUEST
_RESOLVERESOURCEPATHREQUEST.fields_by_name['context'].message_type = _RESOLVERESOURCEPATHREQUEST_CONTEXTENTRY
_RESOLVERESOURCEPATHRESPONSE_CONTEXTENTRY.containing_type = _RESOLVERESOURCEPATHRESPONSE
_RESOLVERESOURCEPATHRESPONSE.fields_by_name['context'].message_type = _RESOLVERESOURCEPATHRESPONSE_CONTEXTENTRY
_PSRESPONSE_INGESTIONSESSION.containing_type = _PSRESPONSE
_PSRESPONSE_PUBLISHSESSION.containing_type = _PSRESPONSE
_PSRESPONSE.fields_by_name['ingestion_sessions'].message_type = _PSRESPONSE_INGESTIONSESSION
_PSRESPONSE.fields_by_name['publish_sessions'].message_type = _PSRESPONSE_PUBLISHSESSION
_LISTINGESTIONSESSIONSREQUEST.fields_by_name['since'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTINGESTIONSESSIONSREQUEST.fields_by_name['until'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTINGESTIONSESSIONSRESPONSE_INGESTIONSESSION.containing_type = _LISTINGESTIONSESSIONSRESPONSE
_LISTINGESTIONSESSIONSRESPONSE.fields_by_name['ingestion_sessions'].message_type = _LISTINGESTIONSESSIONSRESPONSE_INGESTIONSESSION
_APPLYPERSPECTIVEREQUEST.fields_by_name['perspective'].message_type = _PERSPECTIVE
_LISTPERSPECTIVEREQUEST.fields_by_name['pagination_options'].message_type = _PAGINATIONOPTIONS
_LISTPERSPECTIVERESPONSE.fields_by_name['perspectives'].message_type = _PERSPECTIVE
_EVENTMETADATA.fields_by_name['occurred_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PUBLISHRESOURCEFAILEDEVENT.fields_by_name['metadata'].message_type = _EVENTMETADATA
_PUBLISHRESOURCECOMPLETEEVENT.fields_by_name['metadata'].message_type = _EVENTMETADATA
_EXECUTEINGESTIONSESSIONCOMPLETEEVENT.fields_by_name['metadata'].message_type = _EVENTMETADATA
_EXECUTEINGESTIONSESSIONCOMPLETEEVENT.fields_by_name['request'].message_type = _EXECUTEINGESTIONSESSIONREQUEST
_EXECUTEINGESTIONSESSIONFAILEDEVENT.fields_by_name['metadata'].message_type = _EVENTMETADATA
_EXECUTEINGESTIONSESSIONFAILEDEVENT.fields_by_name['request'].message_type = _EXECUTEINGESTIONSESSIONREQUEST
_INGESTIONSESSIONCREATEDEVENT.fields_by_name['metadata'].message_type = _EVENTMETADATA
_INGESTIONSESSIONCREATEDEVENT.fields_by_name['request'].message_type = _CREATEINGESTIONSESSIONREQUEST
_RESOLVEUNIVERSEREQUEST.fields_by_name['universe'].message_type = _UNIVERSE
_RESOLVEUNIVERSEREQUEST.fields_by_name['effective_dating'].message_type = _TIMESTAMPRANGE
_RESOLVEUNIVERSEGROUPENTRY.fields_by_name['entries'].message_type = _RESOLVEUNIVERSEENTRY
_RESOLVEUNIVERSEENTRY.fields_by_name['data_type'].enum_type = _RESOLVEUNIVERSEENTRY_DATATYPE
_RESOLVEUNIVERSEENTRY_DATATYPE.containing_type = _RESOLVEUNIVERSEENTRY
_ERRORENTRY_CONTEXTENTRY.containing_type = _ERRORENTRY
_ERRORENTRY.fields_by_name['occurred_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ERRORENTRY.fields_by_name['context'].message_type = _ERRORENTRY_CONTEXTENTRY
DESCRIPTOR.message_types_by_name['TimestampRange'] = _TIMESTAMPRANGE
DESCRIPTOR.message_types_by_name['PaginationOptions'] = _PAGINATIONOPTIONS
DESCRIPTOR.message_types_by_name['Universe'] = _UNIVERSE
DESCRIPTOR.message_types_by_name['Perspective'] = _PERSPECTIVE
DESCRIPTOR.message_types_by_name['PerspectivePermanentAttribute'] = _PERSPECTIVEPERMANENTATTRIBUTE
DESCRIPTOR.message_types_by_name['MigrationRequest'] = _MIGRATIONREQUEST
DESCRIPTOR.message_types_by_name['MigrationStatusRequest'] = _MIGRATIONSTATUSREQUEST
DESCRIPTOR.message_types_by_name['MigrationResponse'] = _MIGRATIONRESPONSE
DESCRIPTOR.message_types_by_name['ResolveResourceRequest'] = _RESOLVERESOURCEREQUEST
DESCRIPTOR.message_types_by_name['ResolveResourceMappingContextEntry'] = _RESOLVERESOURCEMAPPINGCONTEXTENTRY
DESCRIPTOR.message_types_by_name['ResolveResourceResponse'] = _RESOLVERESOURCERESPONSE
DESCRIPTOR.message_types_by_name['ResolveResourceIntentResponse'] = _RESOLVERESOURCEINTENTRESPONSE
DESCRIPTOR.message_types_by_name['ResolveResourceIntentMapping'] = _RESOLVERESOURCEINTENTMAPPING
DESCRIPTOR.message_types_by_name['ColumnEntry'] = _COLUMNENTRY
DESCRIPTOR.message_types_by_name['ResolveResourceIntentMappingClause'] = _RESOLVERESOURCEINTENTMAPPINGCLAUSE
DESCRIPTOR.message_types_by_name['ConfigurationEntry'] = _CONFIGURATIONENTRY
DESCRIPTOR.message_types_by_name['ApplyLandingZoneRequest'] = _APPLYLANDINGZONEREQUEST
DESCRIPTOR.message_types_by_name['ApplyLandingZoneResponse'] = _APPLYLANDINGZONERESPONSE
DESCRIPTOR.message_types_by_name['RetractLandingZoneRequest'] = _RETRACTLANDINGZONEREQUEST
DESCRIPTOR.message_types_by_name['RetractLandingZoneResponse'] = _RETRACTLANDINGZONERESPONSE
DESCRIPTOR.message_types_by_name['ListLandingZoneRequest'] = _LISTLANDINGZONEREQUEST
DESCRIPTOR.message_types_by_name['ListLandingZoneResponse'] = _LISTLANDINGZONERESPONSE
DESCRIPTOR.message_types_by_name['EnableLandingZoneRequest'] = _ENABLELANDINGZONEREQUEST
DESCRIPTOR.message_types_by_name['EnableLandingZoneResponse'] = _ENABLELANDINGZONERESPONSE
DESCRIPTOR.message_types_by_name['DisableLandingZoneRequest'] = _DISABLELANDINGZONEREQUEST
DESCRIPTOR.message_types_by_name['DisableLandingZoneResponse'] = _DISABLELANDINGZONERESPONSE
DESCRIPTOR.message_types_by_name['SchemaAttribute'] = _SCHEMAATTRIBUTE
DESCRIPTOR.message_types_by_name['Schema'] = _SCHEMA
DESCRIPTOR.message_types_by_name['EntityType'] = _ENTITYTYPE
DESCRIPTOR.message_types_by_name['RelationshipType'] = _RELATIONSHIPTYPE
DESCRIPTOR.message_types_by_name['ApplyOntologyRequest'] = _APPLYONTOLOGYREQUEST
DESCRIPTOR.message_types_by_name['ApplyOntologyResponse'] = _APPLYONTOLOGYRESPONSE
DESCRIPTOR.message_types_by_name['RetractOntologyRequest'] = _RETRACTONTOLOGYREQUEST
DESCRIPTOR.message_types_by_name['RetractOntologyResponse'] = _RETRACTONTOLOGYRESPONSE
DESCRIPTOR.message_types_by_name['ListOntologyRequest'] = _LISTONTOLOGYREQUEST
DESCRIPTOR.message_types_by_name['ListOntologyResponse'] = _LISTONTOLOGYRESPONSE
DESCRIPTOR.message_types_by_name['ExecutionEngine'] = _EXECUTIONENGINE
DESCRIPTOR.message_types_by_name['ApplyExecutionEngineRequest'] = _APPLYEXECUTIONENGINEREQUEST
DESCRIPTOR.message_types_by_name['ApplyExecutionEngineResponse'] = _APPLYEXECUTIONENGINERESPONSE
DESCRIPTOR.message_types_by_name['RetractExecutionEngineRequest'] = _RETRACTEXECUTIONENGINEREQUEST
DESCRIPTOR.message_types_by_name['RetractExecutionEngineResponse'] = _RETRACTEXECUTIONENGINERESPONSE
DESCRIPTOR.message_types_by_name['ListExecutionEngineRequest'] = _LISTEXECUTIONENGINEREQUEST
DESCRIPTOR.message_types_by_name['ListExecutionEngineResponse'] = _LISTEXECUTIONENGINERESPONSE
DESCRIPTOR.message_types_by_name['EnableExecutionEngineRequest'] = _ENABLEEXECUTIONENGINEREQUEST
DESCRIPTOR.message_types_by_name['EnableExecutionEngineResponse'] = _ENABLEEXECUTIONENGINERESPONSE
DESCRIPTOR.message_types_by_name['DisableExecutionEngineRequest'] = _DISABLEEXECUTIONENGINEREQUEST
DESCRIPTOR.message_types_by_name['DisableExecutionEngineResponse'] = _DISABLEEXECUTIONENGINERESPONSE
DESCRIPTOR.message_types_by_name['Warehouse'] = _WAREHOUSE
DESCRIPTOR.message_types_by_name['ApplyWarehouseRequest'] = _APPLYWAREHOUSEREQUEST
DESCRIPTOR.message_types_by_name['ApplyWarehouseResponse'] = _APPLYWAREHOUSERESPONSE
DESCRIPTOR.message_types_by_name['RetractWarehouseRequest'] = _RETRACTWAREHOUSEREQUEST
DESCRIPTOR.message_types_by_name['RetractWarehouseResponse'] = _RETRACTWAREHOUSERESPONSE
DESCRIPTOR.message_types_by_name['ListWarehouseRequest'] = _LISTWAREHOUSEREQUEST
DESCRIPTOR.message_types_by_name['ListWarehouseResponse'] = _LISTWAREHOUSERESPONSE
DESCRIPTOR.message_types_by_name['EnableWarehouseRequest'] = _ENABLEWAREHOUSEREQUEST
DESCRIPTOR.message_types_by_name['EnableWarehouseResponse'] = _ENABLEWAREHOUSERESPONSE
DESCRIPTOR.message_types_by_name['DisableWarehouseRequest'] = _DISABLEWAREHOUSEREQUEST
DESCRIPTOR.message_types_by_name['DisableWarehouseResponse'] = _DISABLEWAREHOUSERESPONSE
DESCRIPTOR.message_types_by_name['ApplyMasteringEngineRequest'] = _APPLYMASTERINGENGINEREQUEST
DESCRIPTOR.message_types_by_name['ApplyMasteringEngineResponse'] = _APPLYMASTERINGENGINERESPONSE
DESCRIPTOR.message_types_by_name['RetractMasteringEngineRequest'] = _RETRACTMASTERINGENGINEREQUEST
DESCRIPTOR.message_types_by_name['RetractMasteringEngineResponse'] = _RETRACTMASTERINGENGINERESPONSE
DESCRIPTOR.message_types_by_name['ResourceStoreContextEntry'] = _RESOURCESTORECONTEXTENTRY
DESCRIPTOR.message_types_by_name['ResourceStore'] = _RESOURCESTORE
DESCRIPTOR.message_types_by_name['ResourcePointerEntry'] = _RESOURCEPOINTERENTRY
DESCRIPTOR.message_types_by_name['ResourceMetadataEntry'] = _RESOURCEMETADATAENTRY
DESCRIPTOR.message_types_by_name['ResourceContextPointerEntry'] = _RESOURCECONTEXTPOINTERENTRY
DESCRIPTOR.message_types_by_name['ResourceDating'] = _RESOURCEDATING
DESCRIPTOR.message_types_by_name['ResourceHashing'] = _RESOURCEHASHING
DESCRIPTOR.message_types_by_name['ResourcePublishing'] = _RESOURCEPUBLISHING
DESCRIPTOR.message_types_by_name['ResourceRetraction'] = _RESOURCERETRACTION
DESCRIPTOR.message_types_by_name['ResourceContext'] = _RESOURCECONTEXT
DESCRIPTOR.message_types_by_name['Resource'] = _RESOURCE
DESCRIPTOR.message_types_by_name['Adapter'] = _ADAPTER
DESCRIPTOR.message_types_by_name['AdapterPersistence'] = _ADAPTERPERSISTENCE
DESCRIPTOR.message_types_by_name['AdapterPersistenceEntry'] = _ADAPTERPERSISTENCEENTRY
DESCRIPTOR.message_types_by_name['AdapterMetadata'] = _ADAPTERMETADATA
DESCRIPTOR.message_types_by_name['Migration'] = _MIGRATION
DESCRIPTOR.message_types_by_name['ApplyAdapterRequest'] = _APPLYADAPTERREQUEST
DESCRIPTOR.message_types_by_name['ApplyAdapterResponse'] = _APPLYADAPTERRESPONSE
DESCRIPTOR.message_types_by_name['RetractAdapterRequest'] = _RETRACTADAPTERREQUEST
DESCRIPTOR.message_types_by_name['RetractAdapterResponse'] = _RETRACTADAPTERRESPONSE
DESCRIPTOR.message_types_by_name['ListAdapterRequest'] = _LISTADAPTERREQUEST
DESCRIPTOR.message_types_by_name['ListAdapterResponse'] = _LISTADAPTERRESPONSE
DESCRIPTOR.message_types_by_name['EnableAdapterRequest'] = _ENABLEADAPTERREQUEST
DESCRIPTOR.message_types_by_name['EnableAdapterResponse'] = _ENABLEADAPTERRESPONSE
DESCRIPTOR.message_types_by_name['DisableAdapterRequest'] = _DISABLEADAPTERREQUEST
DESCRIPTOR.message_types_by_name['DisableAdapterResponse'] = _DISABLEADAPTERRESPONSE
DESCRIPTOR.message_types_by_name['ApplyResourceRequest'] = _APPLYRESOURCEREQUEST
DESCRIPTOR.message_types_by_name['ApplyResourceResponse'] = _APPLYRESOURCERESPONSE
DESCRIPTOR.message_types_by_name['RetractResourceRequest'] = _RETRACTRESOURCEREQUEST
DESCRIPTOR.message_types_by_name['RetractResourceResponse'] = _RETRACTRESOURCERESPONSE
DESCRIPTOR.message_types_by_name['ListResourceRequest'] = _LISTRESOURCEREQUEST
DESCRIPTOR.message_types_by_name['ListResourceResponse'] = _LISTRESOURCERESPONSE
DESCRIPTOR.message_types_by_name['EnableResourceRequest'] = _ENABLERESOURCEREQUEST
DESCRIPTOR.message_types_by_name['EnableResourceResponse'] = _ENABLERESOURCERESPONSE
DESCRIPTOR.message_types_by_name['DisableResourceRequest'] = _DISABLERESOURCEREQUEST
DESCRIPTOR.message_types_by_name['DisableResourceResponse'] = _DISABLERESOURCERESPONSE
DESCRIPTOR.message_types_by_name['AdapterParameters'] = _ADAPTERPARAMETERS
DESCRIPTOR.message_types_by_name['AdapterExecutionContextEntry'] = _ADAPTEREXECUTIONCONTEXTENTRY
DESCRIPTOR.message_types_by_name['AdapterExecutionProfile'] = _ADAPTEREXECUTIONPROFILE
DESCRIPTOR.message_types_by_name['CreateIngestionSessionRequest'] = _CREATEINGESTIONSESSIONREQUEST
DESCRIPTOR.message_types_by_name['CreateIngestionSessionResponse'] = _CREATEINGESTIONSESSIONRESPONSE
DESCRIPTOR.message_types_by_name['ExecuteIngestionSessionRequest'] = _EXECUTEINGESTIONSESSIONREQUEST
DESCRIPTOR.message_types_by_name['GetIngestionSessionRequest'] = _GETINGESTIONSESSIONREQUEST
DESCRIPTOR.message_types_by_name['IngestionSessionStatusResponse'] = _INGESTIONSESSIONSTATUSRESPONSE
DESCRIPTOR.message_types_by_name['MasterEntity'] = _MASTERENTITY
DESCRIPTOR.message_types_by_name['MasterIngestionSessionRequest'] = _MASTERINGESTIONSESSIONREQUEST
DESCRIPTOR.message_types_by_name['MasterIngestionSessionResponse'] = _MASTERINGESTIONSESSIONRESPONSE
DESCRIPTOR.message_types_by_name['PublishIngestionSessionResourceRequest'] = _PUBLISHINGESTIONSESSIONRESOURCEREQUEST
DESCRIPTOR.message_types_by_name['PublishIngestionSessionResourceResponse'] = _PUBLISHINGESTIONSESSIONRESOURCERESPONSE
DESCRIPTOR.message_types_by_name['PublishIngestionSessionResourceStatusRequest'] = _PUBLISHINGESTIONSESSIONRESOURCESTATUSREQUEST
DESCRIPTOR.message_types_by_name['PublishIngestionSessionStatusResponse'] = _PUBLISHINGESTIONSESSIONSTATUSRESPONSE
DESCRIPTOR.message_types_by_name['RetractPublishingSessionRequest'] = _RETRACTPUBLISHINGSESSIONREQUEST
DESCRIPTOR.message_types_by_name['RetractPublishingSessionResponse'] = _RETRACTPUBLISHINGSESSIONRESPONSE
DESCRIPTOR.message_types_by_name['RetractMasteringSessionRequest'] = _RETRACTMASTERINGSESSIONREQUEST
DESCRIPTOR.message_types_by_name['RetractMasteringSessionResponse'] = _RETRACTMASTERINGSESSIONRESPONSE
DESCRIPTOR.message_types_by_name['HaltIncrementalMasteringRequest'] = _HALTINCREMENTALMASTERINGREQUEST
DESCRIPTOR.message_types_by_name['HaltIncrementalMasteringResponse'] = _HALTINCREMENTALMASTERINGRESPONSE
DESCRIPTOR.message_types_by_name['ResumeIncrementalMasteringRequest'] = _RESUMEINCREMENTALMASTERINGREQUEST
DESCRIPTOR.message_types_by_name['ResumeIncrementalMasteringResponse'] = _RESUMEINCREMENTALMASTERINGRESPONSE
DESCRIPTOR.message_types_by_name['ResolveResourcePathRequest'] = _RESOLVERESOURCEPATHREQUEST
DESCRIPTOR.message_types_by_name['ResolveResourcePathResponse'] = _RESOLVERESOURCEPATHRESPONSE
DESCRIPTOR.message_types_by_name['PSRequest'] = _PSREQUEST
DESCRIPTOR.message_types_by_name['PSResponse'] = _PSRESPONSE
DESCRIPTOR.message_types_by_name['ListIngestionSessionArtifactsRequest'] = _LISTINGESTIONSESSIONARTIFACTSREQUEST
DESCRIPTOR.message_types_by_name['ListIngestionSessionArtifactsResponse'] = _LISTINGESTIONSESSIONARTIFACTSRESPONSE
DESCRIPTOR.message_types_by_name['ListIngestionSessionsRequest'] = _LISTINGESTIONSESSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListIngestionSessionsResponse'] = _LISTINGESTIONSESSIONSRESPONSE
DESCRIPTOR.message_types_by_name['ApplyPerspectiveRequest'] = _APPLYPERSPECTIVEREQUEST
DESCRIPTOR.message_types_by_name['ApplyPerspectiveResponse'] = _APPLYPERSPECTIVERESPONSE
DESCRIPTOR.message_types_by_name['RetractPerspectiveRequest'] = _RETRACTPERSPECTIVEREQUEST
DESCRIPTOR.message_types_by_name['RetractPerspectiveResponse'] = _RETRACTPERSPECTIVERESPONSE
DESCRIPTOR.message_types_by_name['ListPerspectiveRequest'] = _LISTPERSPECTIVEREQUEST
DESCRIPTOR.message_types_by_name['ListPerspectiveResponse'] = _LISTPERSPECTIVERESPONSE
DESCRIPTOR.message_types_by_name['EventMetadata'] = _EVENTMETADATA
DESCRIPTOR.message_types_by_name['PublishResourceFailedEvent'] = _PUBLISHRESOURCEFAILEDEVENT
DESCRIPTOR.message_types_by_name['PublishResourceCompleteEvent'] = _PUBLISHRESOURCECOMPLETEEVENT
DESCRIPTOR.message_types_by_name['ExecuteIngestionSessionCompleteEvent'] = _EXECUTEINGESTIONSESSIONCOMPLETEEVENT
DESCRIPTOR.message_types_by_name['ExecuteIngestionSessionFailedEvent'] = _EXECUTEINGESTIONSESSIONFAILEDEVENT
DESCRIPTOR.message_types_by_name['IngestionSessionCreatedEvent'] = _INGESTIONSESSIONCREATEDEVENT
DESCRIPTOR.message_types_by_name['ResolveUniverseRequest'] = _RESOLVEUNIVERSEREQUEST
DESCRIPTOR.message_types_by_name['ResolveUniverseGroupEntry'] = _RESOLVEUNIVERSEGROUPENTRY
DESCRIPTOR.message_types_by_name['ResolveUniverseEntry'] = _RESOLVEUNIVERSEENTRY
DESCRIPTOR.message_types_by_name['ErrorEntry'] = _ERRORENTRY
DESCRIPTOR.enum_types_by_name['ComparisonType'] = _COMPARISONTYPE
DESCRIPTOR.enum_types_by_name['AttributeType'] = _ATTRIBUTETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TimestampRange = _reflection.GeneratedProtocolMessageType('TimestampRange', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPRANGE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.TimestampRange)
  })
_sym_db.RegisterMessage(TimestampRange)

PaginationOptions = _reflection.GeneratedProtocolMessageType('PaginationOptions', (_message.Message,), {
  'DESCRIPTOR' : _PAGINATIONOPTIONS,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PaginationOptions)
  })
_sym_db.RegisterMessage(PaginationOptions)

Universe = _reflection.GeneratedProtocolMessageType('Universe', (_message.Message,), {
  'DESCRIPTOR' : _UNIVERSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.Universe)
  })
_sym_db.RegisterMessage(Universe)

Perspective = _reflection.GeneratedProtocolMessageType('Perspective', (_message.Message,), {
  'DESCRIPTOR' : _PERSPECTIVE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.Perspective)
  })
_sym_db.RegisterMessage(Perspective)

PerspectivePermanentAttribute = _reflection.GeneratedProtocolMessageType('PerspectivePermanentAttribute', (_message.Message,), {
  'DESCRIPTOR' : _PERSPECTIVEPERMANENTATTRIBUTE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PerspectivePermanentAttribute)
  })
_sym_db.RegisterMessage(PerspectivePermanentAttribute)

MigrationRequest = _reflection.GeneratedProtocolMessageType('MigrationRequest', (_message.Message,), {

  'Force' : _reflection.GeneratedProtocolMessageType('Force', (_message.Message,), {
    'DESCRIPTOR' : _MIGRATIONREQUEST_FORCE,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.MigrationRequest.Force)
    })
  ,
  'DESCRIPTOR' : _MIGRATIONREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.MigrationRequest)
  })
_sym_db.RegisterMessage(MigrationRequest)
_sym_db.RegisterMessage(MigrationRequest.Force)

MigrationStatusRequest = _reflection.GeneratedProtocolMessageType('MigrationStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _MIGRATIONSTATUSREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.MigrationStatusRequest)
  })
_sym_db.RegisterMessage(MigrationStatusRequest)

MigrationResponse = _reflection.GeneratedProtocolMessageType('MigrationResponse', (_message.Message,), {
  'DESCRIPTOR' : _MIGRATIONRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.MigrationResponse)
  })
_sym_db.RegisterMessage(MigrationResponse)

ResolveResourceRequest = _reflection.GeneratedProtocolMessageType('ResolveResourceRequest', (_message.Message,), {

  'PointersEntry' : _reflection.GeneratedProtocolMessageType('PointersEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOLVERESOURCEREQUEST_POINTERSENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceRequest.PointersEntry)
    })
  ,
  'DESCRIPTOR' : _RESOLVERESOURCEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceRequest)
  })
_sym_db.RegisterMessage(ResolveResourceRequest)
_sym_db.RegisterMessage(ResolveResourceRequest.PointersEntry)

ResolveResourceMappingContextEntry = _reflection.GeneratedProtocolMessageType('ResolveResourceMappingContextEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOLVERESOURCEMAPPINGCONTEXTENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceMappingContextEntry)
  })
_sym_db.RegisterMessage(ResolveResourceMappingContextEntry)

ResolveResourceResponse = _reflection.GeneratedProtocolMessageType('ResolveResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESOLVERESOURCERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceResponse)
  })
_sym_db.RegisterMessage(ResolveResourceResponse)

ResolveResourceIntentResponse = _reflection.GeneratedProtocolMessageType('ResolveResourceIntentResponse', (_message.Message,), {

  'ContextEntry' : _reflection.GeneratedProtocolMessageType('ContextEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOLVERESOURCEINTENTRESPONSE_CONTEXTENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceIntentResponse.ContextEntry)
    })
  ,
  'DESCRIPTOR' : _RESOLVERESOURCEINTENTRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceIntentResponse)
  })
_sym_db.RegisterMessage(ResolveResourceIntentResponse)
_sym_db.RegisterMessage(ResolveResourceIntentResponse.ContextEntry)

ResolveResourceIntentMapping = _reflection.GeneratedProtocolMessageType('ResolveResourceIntentMapping', (_message.Message,), {
  'DESCRIPTOR' : _RESOLVERESOURCEINTENTMAPPING,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceIntentMapping)
  })
_sym_db.RegisterMessage(ResolveResourceIntentMapping)

ColumnEntry = _reflection.GeneratedProtocolMessageType('ColumnEntry', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ColumnEntry)
  })
_sym_db.RegisterMessage(ColumnEntry)

ResolveResourceIntentMappingClause = _reflection.GeneratedProtocolMessageType('ResolveResourceIntentMappingClause', (_message.Message,), {
  'DESCRIPTOR' : _RESOLVERESOURCEINTENTMAPPINGCLAUSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourceIntentMappingClause)
  })
_sym_db.RegisterMessage(ResolveResourceIntentMappingClause)

ConfigurationEntry = _reflection.GeneratedProtocolMessageType('ConfigurationEntry', (_message.Message,), {
  'DESCRIPTOR' : _CONFIGURATIONENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ConfigurationEntry)
  })
_sym_db.RegisterMessage(ConfigurationEntry)

ApplyLandingZoneRequest = _reflection.GeneratedProtocolMessageType('ApplyLandingZoneRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYLANDINGZONEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyLandingZoneRequest)
  })
_sym_db.RegisterMessage(ApplyLandingZoneRequest)

ApplyLandingZoneResponse = _reflection.GeneratedProtocolMessageType('ApplyLandingZoneResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYLANDINGZONERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyLandingZoneResponse)
  })
_sym_db.RegisterMessage(ApplyLandingZoneResponse)

RetractLandingZoneRequest = _reflection.GeneratedProtocolMessageType('RetractLandingZoneRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTLANDINGZONEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractLandingZoneRequest)
  })
_sym_db.RegisterMessage(RetractLandingZoneRequest)

RetractLandingZoneResponse = _reflection.GeneratedProtocolMessageType('RetractLandingZoneResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTLANDINGZONERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractLandingZoneResponse)
  })
_sym_db.RegisterMessage(RetractLandingZoneResponse)

ListLandingZoneRequest = _reflection.GeneratedProtocolMessageType('ListLandingZoneRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTLANDINGZONEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListLandingZoneRequest)
  })
_sym_db.RegisterMessage(ListLandingZoneRequest)

ListLandingZoneResponse = _reflection.GeneratedProtocolMessageType('ListLandingZoneResponse', (_message.Message,), {

  'LandingZoneDetails' : _reflection.GeneratedProtocolMessageType('LandingZoneDetails', (_message.Message,), {
    'DESCRIPTOR' : _LISTLANDINGZONERESPONSE_LANDINGZONEDETAILS,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ListLandingZoneResponse.LandingZoneDetails)
    })
  ,
  'DESCRIPTOR' : _LISTLANDINGZONERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListLandingZoneResponse)
  })
_sym_db.RegisterMessage(ListLandingZoneResponse)
_sym_db.RegisterMessage(ListLandingZoneResponse.LandingZoneDetails)

EnableLandingZoneRequest = _reflection.GeneratedProtocolMessageType('EnableLandingZoneRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENABLELANDINGZONEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableLandingZoneRequest)
  })
_sym_db.RegisterMessage(EnableLandingZoneRequest)

EnableLandingZoneResponse = _reflection.GeneratedProtocolMessageType('EnableLandingZoneResponse', (_message.Message,), {
  'DESCRIPTOR' : _ENABLELANDINGZONERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableLandingZoneResponse)
  })
_sym_db.RegisterMessage(EnableLandingZoneResponse)

DisableLandingZoneRequest = _reflection.GeneratedProtocolMessageType('DisableLandingZoneRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISABLELANDINGZONEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableLandingZoneRequest)
  })
_sym_db.RegisterMessage(DisableLandingZoneRequest)

DisableLandingZoneResponse = _reflection.GeneratedProtocolMessageType('DisableLandingZoneResponse', (_message.Message,), {
  'DESCRIPTOR' : _DISABLELANDINGZONERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableLandingZoneResponse)
  })
_sym_db.RegisterMessage(DisableLandingZoneResponse)

SchemaAttribute = _reflection.GeneratedProtocolMessageType('SchemaAttribute', (_message.Message,), {
  'DESCRIPTOR' : _SCHEMAATTRIBUTE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.SchemaAttribute)
  })
_sym_db.RegisterMessage(SchemaAttribute)

Schema = _reflection.GeneratedProtocolMessageType('Schema', (_message.Message,), {
  'DESCRIPTOR' : _SCHEMA,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.Schema)
  })
_sym_db.RegisterMessage(Schema)

EntityType = _reflection.GeneratedProtocolMessageType('EntityType', (_message.Message,), {
  'DESCRIPTOR' : _ENTITYTYPE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EntityType)
  })
_sym_db.RegisterMessage(EntityType)

RelationshipType = _reflection.GeneratedProtocolMessageType('RelationshipType', (_message.Message,), {
  'DESCRIPTOR' : _RELATIONSHIPTYPE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RelationshipType)
  })
_sym_db.RegisterMessage(RelationshipType)

ApplyOntologyRequest = _reflection.GeneratedProtocolMessageType('ApplyOntologyRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYONTOLOGYREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyOntologyRequest)
  })
_sym_db.RegisterMessage(ApplyOntologyRequest)

ApplyOntologyResponse = _reflection.GeneratedProtocolMessageType('ApplyOntologyResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYONTOLOGYRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyOntologyResponse)
  })
_sym_db.RegisterMessage(ApplyOntologyResponse)

RetractOntologyRequest = _reflection.GeneratedProtocolMessageType('RetractOntologyRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTONTOLOGYREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractOntologyRequest)
  })
_sym_db.RegisterMessage(RetractOntologyRequest)

RetractOntologyResponse = _reflection.GeneratedProtocolMessageType('RetractOntologyResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTONTOLOGYRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractOntologyResponse)
  })
_sym_db.RegisterMessage(RetractOntologyResponse)

ListOntologyRequest = _reflection.GeneratedProtocolMessageType('ListOntologyRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTONTOLOGYREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListOntologyRequest)
  })
_sym_db.RegisterMessage(ListOntologyRequest)

ListOntologyResponse = _reflection.GeneratedProtocolMessageType('ListOntologyResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTONTOLOGYRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListOntologyResponse)
  })
_sym_db.RegisterMessage(ListOntologyResponse)

ExecutionEngine = _reflection.GeneratedProtocolMessageType('ExecutionEngine', (_message.Message,), {
  'DESCRIPTOR' : _EXECUTIONENGINE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ExecutionEngine)
  })
_sym_db.RegisterMessage(ExecutionEngine)

ApplyExecutionEngineRequest = _reflection.GeneratedProtocolMessageType('ApplyExecutionEngineRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYEXECUTIONENGINEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyExecutionEngineRequest)
  })
_sym_db.RegisterMessage(ApplyExecutionEngineRequest)

ApplyExecutionEngineResponse = _reflection.GeneratedProtocolMessageType('ApplyExecutionEngineResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYEXECUTIONENGINERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyExecutionEngineResponse)
  })
_sym_db.RegisterMessage(ApplyExecutionEngineResponse)

RetractExecutionEngineRequest = _reflection.GeneratedProtocolMessageType('RetractExecutionEngineRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTEXECUTIONENGINEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractExecutionEngineRequest)
  })
_sym_db.RegisterMessage(RetractExecutionEngineRequest)

RetractExecutionEngineResponse = _reflection.GeneratedProtocolMessageType('RetractExecutionEngineResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTEXECUTIONENGINERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractExecutionEngineResponse)
  })
_sym_db.RegisterMessage(RetractExecutionEngineResponse)

ListExecutionEngineRequest = _reflection.GeneratedProtocolMessageType('ListExecutionEngineRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTEXECUTIONENGINEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListExecutionEngineRequest)
  })
_sym_db.RegisterMessage(ListExecutionEngineRequest)

ListExecutionEngineResponse = _reflection.GeneratedProtocolMessageType('ListExecutionEngineResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTEXECUTIONENGINERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListExecutionEngineResponse)
  })
_sym_db.RegisterMessage(ListExecutionEngineResponse)

EnableExecutionEngineRequest = _reflection.GeneratedProtocolMessageType('EnableExecutionEngineRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENABLEEXECUTIONENGINEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableExecutionEngineRequest)
  })
_sym_db.RegisterMessage(EnableExecutionEngineRequest)

EnableExecutionEngineResponse = _reflection.GeneratedProtocolMessageType('EnableExecutionEngineResponse', (_message.Message,), {
  'DESCRIPTOR' : _ENABLEEXECUTIONENGINERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableExecutionEngineResponse)
  })
_sym_db.RegisterMessage(EnableExecutionEngineResponse)

DisableExecutionEngineRequest = _reflection.GeneratedProtocolMessageType('DisableExecutionEngineRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISABLEEXECUTIONENGINEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableExecutionEngineRequest)
  })
_sym_db.RegisterMessage(DisableExecutionEngineRequest)

DisableExecutionEngineResponse = _reflection.GeneratedProtocolMessageType('DisableExecutionEngineResponse', (_message.Message,), {
  'DESCRIPTOR' : _DISABLEEXECUTIONENGINERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableExecutionEngineResponse)
  })
_sym_db.RegisterMessage(DisableExecutionEngineResponse)

Warehouse = _reflection.GeneratedProtocolMessageType('Warehouse', (_message.Message,), {
  'DESCRIPTOR' : _WAREHOUSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.Warehouse)
  })
_sym_db.RegisterMessage(Warehouse)

ApplyWarehouseRequest = _reflection.GeneratedProtocolMessageType('ApplyWarehouseRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYWAREHOUSEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyWarehouseRequest)
  })
_sym_db.RegisterMessage(ApplyWarehouseRequest)

ApplyWarehouseResponse = _reflection.GeneratedProtocolMessageType('ApplyWarehouseResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYWAREHOUSERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyWarehouseResponse)
  })
_sym_db.RegisterMessage(ApplyWarehouseResponse)

RetractWarehouseRequest = _reflection.GeneratedProtocolMessageType('RetractWarehouseRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTWAREHOUSEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractWarehouseRequest)
  })
_sym_db.RegisterMessage(RetractWarehouseRequest)

RetractWarehouseResponse = _reflection.GeneratedProtocolMessageType('RetractWarehouseResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTWAREHOUSERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractWarehouseResponse)
  })
_sym_db.RegisterMessage(RetractWarehouseResponse)

ListWarehouseRequest = _reflection.GeneratedProtocolMessageType('ListWarehouseRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTWAREHOUSEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListWarehouseRequest)
  })
_sym_db.RegisterMessage(ListWarehouseRequest)

ListWarehouseResponse = _reflection.GeneratedProtocolMessageType('ListWarehouseResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTWAREHOUSERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListWarehouseResponse)
  })
_sym_db.RegisterMessage(ListWarehouseResponse)

EnableWarehouseRequest = _reflection.GeneratedProtocolMessageType('EnableWarehouseRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENABLEWAREHOUSEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableWarehouseRequest)
  })
_sym_db.RegisterMessage(EnableWarehouseRequest)

EnableWarehouseResponse = _reflection.GeneratedProtocolMessageType('EnableWarehouseResponse', (_message.Message,), {
  'DESCRIPTOR' : _ENABLEWAREHOUSERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableWarehouseResponse)
  })
_sym_db.RegisterMessage(EnableWarehouseResponse)

DisableWarehouseRequest = _reflection.GeneratedProtocolMessageType('DisableWarehouseRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISABLEWAREHOUSEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableWarehouseRequest)
  })
_sym_db.RegisterMessage(DisableWarehouseRequest)

DisableWarehouseResponse = _reflection.GeneratedProtocolMessageType('DisableWarehouseResponse', (_message.Message,), {
  'DESCRIPTOR' : _DISABLEWAREHOUSERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableWarehouseResponse)
  })
_sym_db.RegisterMessage(DisableWarehouseResponse)

ApplyMasteringEngineRequest = _reflection.GeneratedProtocolMessageType('ApplyMasteringEngineRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYMASTERINGENGINEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyMasteringEngineRequest)
  })
_sym_db.RegisterMessage(ApplyMasteringEngineRequest)

ApplyMasteringEngineResponse = _reflection.GeneratedProtocolMessageType('ApplyMasteringEngineResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYMASTERINGENGINERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyMasteringEngineResponse)
  })
_sym_db.RegisterMessage(ApplyMasteringEngineResponse)

RetractMasteringEngineRequest = _reflection.GeneratedProtocolMessageType('RetractMasteringEngineRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTMASTERINGENGINEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractMasteringEngineRequest)
  })
_sym_db.RegisterMessage(RetractMasteringEngineRequest)

RetractMasteringEngineResponse = _reflection.GeneratedProtocolMessageType('RetractMasteringEngineResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTMASTERINGENGINERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractMasteringEngineResponse)
  })
_sym_db.RegisterMessage(RetractMasteringEngineResponse)

ResourceStoreContextEntry = _reflection.GeneratedProtocolMessageType('ResourceStoreContextEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCESTORECONTEXTENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceStoreContextEntry)
  })
_sym_db.RegisterMessage(ResourceStoreContextEntry)

ResourceStore = _reflection.GeneratedProtocolMessageType('ResourceStore', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCESTORE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceStore)
  })
_sym_db.RegisterMessage(ResourceStore)

ResourcePointerEntry = _reflection.GeneratedProtocolMessageType('ResourcePointerEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEPOINTERENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourcePointerEntry)
  })
_sym_db.RegisterMessage(ResourcePointerEntry)

ResourceMetadataEntry = _reflection.GeneratedProtocolMessageType('ResourceMetadataEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEMETADATAENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceMetadataEntry)
  })
_sym_db.RegisterMessage(ResourceMetadataEntry)

ResourceContextPointerEntry = _reflection.GeneratedProtocolMessageType('ResourceContextPointerEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCECONTEXTPOINTERENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceContextPointerEntry)
  })
_sym_db.RegisterMessage(ResourceContextPointerEntry)

ResourceDating = _reflection.GeneratedProtocolMessageType('ResourceDating', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEDATING,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceDating)
  })
_sym_db.RegisterMessage(ResourceDating)

ResourceHashing = _reflection.GeneratedProtocolMessageType('ResourceHashing', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEHASHING,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceHashing)
  })
_sym_db.RegisterMessage(ResourceHashing)

ResourcePublishing = _reflection.GeneratedProtocolMessageType('ResourcePublishing', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEPUBLISHING,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourcePublishing)
  })
_sym_db.RegisterMessage(ResourcePublishing)

ResourceRetraction = _reflection.GeneratedProtocolMessageType('ResourceRetraction', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCERETRACTION,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceRetraction)
  })
_sym_db.RegisterMessage(ResourceRetraction)

ResourceContext = _reflection.GeneratedProtocolMessageType('ResourceContext', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCECONTEXT,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResourceContext)
  })
_sym_db.RegisterMessage(ResourceContext)

Resource = _reflection.GeneratedProtocolMessageType('Resource', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.Resource)
  })
_sym_db.RegisterMessage(Resource)

Adapter = _reflection.GeneratedProtocolMessageType('Adapter', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTER,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.Adapter)
  })
_sym_db.RegisterMessage(Adapter)

AdapterPersistence = _reflection.GeneratedProtocolMessageType('AdapterPersistence', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTERPERSISTENCE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.AdapterPersistence)
  })
_sym_db.RegisterMessage(AdapterPersistence)

AdapterPersistenceEntry = _reflection.GeneratedProtocolMessageType('AdapterPersistenceEntry', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTERPERSISTENCEENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.AdapterPersistenceEntry)
  })
_sym_db.RegisterMessage(AdapterPersistenceEntry)

AdapterMetadata = _reflection.GeneratedProtocolMessageType('AdapterMetadata', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTERMETADATA,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.AdapterMetadata)
  })
_sym_db.RegisterMessage(AdapterMetadata)

Migration = _reflection.GeneratedProtocolMessageType('Migration', (_message.Message,), {
  'DESCRIPTOR' : _MIGRATION,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.Migration)
  })
_sym_db.RegisterMessage(Migration)

ApplyAdapterRequest = _reflection.GeneratedProtocolMessageType('ApplyAdapterRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYADAPTERREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyAdapterRequest)
  })
_sym_db.RegisterMessage(ApplyAdapterRequest)

ApplyAdapterResponse = _reflection.GeneratedProtocolMessageType('ApplyAdapterResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYADAPTERRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyAdapterResponse)
  })
_sym_db.RegisterMessage(ApplyAdapterResponse)

RetractAdapterRequest = _reflection.GeneratedProtocolMessageType('RetractAdapterRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTADAPTERREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractAdapterRequest)
  })
_sym_db.RegisterMessage(RetractAdapterRequest)

RetractAdapterResponse = _reflection.GeneratedProtocolMessageType('RetractAdapterResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTADAPTERRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractAdapterResponse)
  })
_sym_db.RegisterMessage(RetractAdapterResponse)

ListAdapterRequest = _reflection.GeneratedProtocolMessageType('ListAdapterRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTADAPTERREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListAdapterRequest)
  })
_sym_db.RegisterMessage(ListAdapterRequest)

ListAdapterResponse = _reflection.GeneratedProtocolMessageType('ListAdapterResponse', (_message.Message,), {

  'AdapterDetails' : _reflection.GeneratedProtocolMessageType('AdapterDetails', (_message.Message,), {
    'DESCRIPTOR' : _LISTADAPTERRESPONSE_ADAPTERDETAILS,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ListAdapterResponse.AdapterDetails)
    })
  ,
  'DESCRIPTOR' : _LISTADAPTERRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListAdapterResponse)
  })
_sym_db.RegisterMessage(ListAdapterResponse)
_sym_db.RegisterMessage(ListAdapterResponse.AdapterDetails)

EnableAdapterRequest = _reflection.GeneratedProtocolMessageType('EnableAdapterRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENABLEADAPTERREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableAdapterRequest)
  })
_sym_db.RegisterMessage(EnableAdapterRequest)

EnableAdapterResponse = _reflection.GeneratedProtocolMessageType('EnableAdapterResponse', (_message.Message,), {
  'DESCRIPTOR' : _ENABLEADAPTERRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableAdapterResponse)
  })
_sym_db.RegisterMessage(EnableAdapterResponse)

DisableAdapterRequest = _reflection.GeneratedProtocolMessageType('DisableAdapterRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISABLEADAPTERREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableAdapterRequest)
  })
_sym_db.RegisterMessage(DisableAdapterRequest)

DisableAdapterResponse = _reflection.GeneratedProtocolMessageType('DisableAdapterResponse', (_message.Message,), {
  'DESCRIPTOR' : _DISABLEADAPTERRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableAdapterResponse)
  })
_sym_db.RegisterMessage(DisableAdapterResponse)

ApplyResourceRequest = _reflection.GeneratedProtocolMessageType('ApplyResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYRESOURCEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyResourceRequest)
  })
_sym_db.RegisterMessage(ApplyResourceRequest)

ApplyResourceResponse = _reflection.GeneratedProtocolMessageType('ApplyResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYRESOURCERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyResourceResponse)
  })
_sym_db.RegisterMessage(ApplyResourceResponse)

RetractResourceRequest = _reflection.GeneratedProtocolMessageType('RetractResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTRESOURCEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractResourceRequest)
  })
_sym_db.RegisterMessage(RetractResourceRequest)

RetractResourceResponse = _reflection.GeneratedProtocolMessageType('RetractResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTRESOURCERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractResourceResponse)
  })
_sym_db.RegisterMessage(RetractResourceResponse)

ListResourceRequest = _reflection.GeneratedProtocolMessageType('ListResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTRESOURCEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListResourceRequest)
  })
_sym_db.RegisterMessage(ListResourceRequest)

ListResourceResponse = _reflection.GeneratedProtocolMessageType('ListResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRESOURCERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListResourceResponse)
  })
_sym_db.RegisterMessage(ListResourceResponse)

EnableResourceRequest = _reflection.GeneratedProtocolMessageType('EnableResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENABLERESOURCEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableResourceRequest)
  })
_sym_db.RegisterMessage(EnableResourceRequest)

EnableResourceResponse = _reflection.GeneratedProtocolMessageType('EnableResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _ENABLERESOURCERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EnableResourceResponse)
  })
_sym_db.RegisterMessage(EnableResourceResponse)

DisableResourceRequest = _reflection.GeneratedProtocolMessageType('DisableResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISABLERESOURCEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableResourceRequest)
  })
_sym_db.RegisterMessage(DisableResourceRequest)

DisableResourceResponse = _reflection.GeneratedProtocolMessageType('DisableResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _DISABLERESOURCERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.DisableResourceResponse)
  })
_sym_db.RegisterMessage(DisableResourceResponse)

AdapterParameters = _reflection.GeneratedProtocolMessageType('AdapterParameters', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTERPARAMETERS,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.AdapterParameters)
  })
_sym_db.RegisterMessage(AdapterParameters)

AdapterExecutionContextEntry = _reflection.GeneratedProtocolMessageType('AdapterExecutionContextEntry', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTEREXECUTIONCONTEXTENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.AdapterExecutionContextEntry)
  })
_sym_db.RegisterMessage(AdapterExecutionContextEntry)

AdapterExecutionProfile = _reflection.GeneratedProtocolMessageType('AdapterExecutionProfile', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTEREXECUTIONPROFILE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.AdapterExecutionProfile)
  })
_sym_db.RegisterMessage(AdapterExecutionProfile)

CreateIngestionSessionRequest = _reflection.GeneratedProtocolMessageType('CreateIngestionSessionRequest', (_message.Message,), {

  'ParametersEntry' : _reflection.GeneratedProtocolMessageType('ParametersEntry', (_message.Message,), {
    'DESCRIPTOR' : _CREATEINGESTIONSESSIONREQUEST_PARAMETERSENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.CreateIngestionSessionRequest.ParametersEntry)
    })
  ,
  'DESCRIPTOR' : _CREATEINGESTIONSESSIONREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.CreateIngestionSessionRequest)
  })
_sym_db.RegisterMessage(CreateIngestionSessionRequest)
_sym_db.RegisterMessage(CreateIngestionSessionRequest.ParametersEntry)

CreateIngestionSessionResponse = _reflection.GeneratedProtocolMessageType('CreateIngestionSessionResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEINGESTIONSESSIONRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.CreateIngestionSessionResponse)
  })
_sym_db.RegisterMessage(CreateIngestionSessionResponse)

ExecuteIngestionSessionRequest = _reflection.GeneratedProtocolMessageType('ExecuteIngestionSessionRequest', (_message.Message,), {

  'ParametersEntry' : _reflection.GeneratedProtocolMessageType('ParametersEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXECUTEINGESTIONSESSIONREQUEST_PARAMETERSENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ExecuteIngestionSessionRequest.ParametersEntry)
    })
  ,
  'DESCRIPTOR' : _EXECUTEINGESTIONSESSIONREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ExecuteIngestionSessionRequest)
  })
_sym_db.RegisterMessage(ExecuteIngestionSessionRequest)
_sym_db.RegisterMessage(ExecuteIngestionSessionRequest.ParametersEntry)

GetIngestionSessionRequest = _reflection.GeneratedProtocolMessageType('GetIngestionSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINGESTIONSESSIONREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.GetIngestionSessionRequest)
  })
_sym_db.RegisterMessage(GetIngestionSessionRequest)

IngestionSessionStatusResponse = _reflection.GeneratedProtocolMessageType('IngestionSessionStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _INGESTIONSESSIONSTATUSRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.IngestionSessionStatusResponse)
  })
_sym_db.RegisterMessage(IngestionSessionStatusResponse)

MasterEntity = _reflection.GeneratedProtocolMessageType('MasterEntity', (_message.Message,), {

  'DescriptorsEntry' : _reflection.GeneratedProtocolMessageType('DescriptorsEntry', (_message.Message,), {
    'DESCRIPTOR' : _MASTERENTITY_DESCRIPTORSENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.MasterEntity.DescriptorsEntry)
    })
  ,
  'DESCRIPTOR' : _MASTERENTITY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.MasterEntity)
  })
_sym_db.RegisterMessage(MasterEntity)
_sym_db.RegisterMessage(MasterEntity.DescriptorsEntry)

MasterIngestionSessionRequest = _reflection.GeneratedProtocolMessageType('MasterIngestionSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _MASTERINGESTIONSESSIONREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.MasterIngestionSessionRequest)
  })
_sym_db.RegisterMessage(MasterIngestionSessionRequest)

MasterIngestionSessionResponse = _reflection.GeneratedProtocolMessageType('MasterIngestionSessionResponse', (_message.Message,), {
  'DESCRIPTOR' : _MASTERINGESTIONSESSIONRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.MasterIngestionSessionResponse)
  })
_sym_db.RegisterMessage(MasterIngestionSessionResponse)

PublishIngestionSessionResourceRequest = _reflection.GeneratedProtocolMessageType('PublishIngestionSessionResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUBLISHINGESTIONSESSIONRESOURCEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PublishIngestionSessionResourceRequest)
  })
_sym_db.RegisterMessage(PublishIngestionSessionResourceRequest)

PublishIngestionSessionResourceResponse = _reflection.GeneratedProtocolMessageType('PublishIngestionSessionResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _PUBLISHINGESTIONSESSIONRESOURCERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PublishIngestionSessionResourceResponse)
  })
_sym_db.RegisterMessage(PublishIngestionSessionResourceResponse)

PublishIngestionSessionResourceStatusRequest = _reflection.GeneratedProtocolMessageType('PublishIngestionSessionResourceStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUBLISHINGESTIONSESSIONRESOURCESTATUSREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PublishIngestionSessionResourceStatusRequest)
  })
_sym_db.RegisterMessage(PublishIngestionSessionResourceStatusRequest)

PublishIngestionSessionStatusResponse = _reflection.GeneratedProtocolMessageType('PublishIngestionSessionStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _PUBLISHINGESTIONSESSIONSTATUSRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PublishIngestionSessionStatusResponse)
  })
_sym_db.RegisterMessage(PublishIngestionSessionStatusResponse)

RetractPublishingSessionRequest = _reflection.GeneratedProtocolMessageType('RetractPublishingSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTPUBLISHINGSESSIONREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractPublishingSessionRequest)
  })
_sym_db.RegisterMessage(RetractPublishingSessionRequest)

RetractPublishingSessionResponse = _reflection.GeneratedProtocolMessageType('RetractPublishingSessionResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTPUBLISHINGSESSIONRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractPublishingSessionResponse)
  })
_sym_db.RegisterMessage(RetractPublishingSessionResponse)

RetractMasteringSessionRequest = _reflection.GeneratedProtocolMessageType('RetractMasteringSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTMASTERINGSESSIONREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractMasteringSessionRequest)
  })
_sym_db.RegisterMessage(RetractMasteringSessionRequest)

RetractMasteringSessionResponse = _reflection.GeneratedProtocolMessageType('RetractMasteringSessionResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTMASTERINGSESSIONRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractMasteringSessionResponse)
  })
_sym_db.RegisterMessage(RetractMasteringSessionResponse)

HaltIncrementalMasteringRequest = _reflection.GeneratedProtocolMessageType('HaltIncrementalMasteringRequest', (_message.Message,), {
  'DESCRIPTOR' : _HALTINCREMENTALMASTERINGREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.HaltIncrementalMasteringRequest)
  })
_sym_db.RegisterMessage(HaltIncrementalMasteringRequest)

HaltIncrementalMasteringResponse = _reflection.GeneratedProtocolMessageType('HaltIncrementalMasteringResponse', (_message.Message,), {
  'DESCRIPTOR' : _HALTINCREMENTALMASTERINGRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.HaltIncrementalMasteringResponse)
  })
_sym_db.RegisterMessage(HaltIncrementalMasteringResponse)

ResumeIncrementalMasteringRequest = _reflection.GeneratedProtocolMessageType('ResumeIncrementalMasteringRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESUMEINCREMENTALMASTERINGREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResumeIncrementalMasteringRequest)
  })
_sym_db.RegisterMessage(ResumeIncrementalMasteringRequest)

ResumeIncrementalMasteringResponse = _reflection.GeneratedProtocolMessageType('ResumeIncrementalMasteringResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESUMEINCREMENTALMASTERINGRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResumeIncrementalMasteringResponse)
  })
_sym_db.RegisterMessage(ResumeIncrementalMasteringResponse)

ResolveResourcePathRequest = _reflection.GeneratedProtocolMessageType('ResolveResourcePathRequest', (_message.Message,), {

  'ContextEntry' : _reflection.GeneratedProtocolMessageType('ContextEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOLVERESOURCEPATHREQUEST_CONTEXTENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourcePathRequest.ContextEntry)
    })
  ,
  'DESCRIPTOR' : _RESOLVERESOURCEPATHREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourcePathRequest)
  })
_sym_db.RegisterMessage(ResolveResourcePathRequest)
_sym_db.RegisterMessage(ResolveResourcePathRequest.ContextEntry)

ResolveResourcePathResponse = _reflection.GeneratedProtocolMessageType('ResolveResourcePathResponse', (_message.Message,), {

  'ContextEntry' : _reflection.GeneratedProtocolMessageType('ContextEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOLVERESOURCEPATHRESPONSE_CONTEXTENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourcePathResponse.ContextEntry)
    })
  ,
  'DESCRIPTOR' : _RESOLVERESOURCEPATHRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveResourcePathResponse)
  })
_sym_db.RegisterMessage(ResolveResourcePathResponse)
_sym_db.RegisterMessage(ResolveResourcePathResponse.ContextEntry)

PSRequest = _reflection.GeneratedProtocolMessageType('PSRequest', (_message.Message,), {
  'DESCRIPTOR' : _PSREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PSRequest)
  })
_sym_db.RegisterMessage(PSRequest)

PSResponse = _reflection.GeneratedProtocolMessageType('PSResponse', (_message.Message,), {

  'IngestionSession' : _reflection.GeneratedProtocolMessageType('IngestionSession', (_message.Message,), {
    'DESCRIPTOR' : _PSRESPONSE_INGESTIONSESSION,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.PSResponse.IngestionSession)
    })
  ,

  'PublishSession' : _reflection.GeneratedProtocolMessageType('PublishSession', (_message.Message,), {
    'DESCRIPTOR' : _PSRESPONSE_PUBLISHSESSION,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.PSResponse.PublishSession)
    })
  ,
  'DESCRIPTOR' : _PSRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PSResponse)
  })
_sym_db.RegisterMessage(PSResponse)
_sym_db.RegisterMessage(PSResponse.IngestionSession)
_sym_db.RegisterMessage(PSResponse.PublishSession)

ListIngestionSessionArtifactsRequest = _reflection.GeneratedProtocolMessageType('ListIngestionSessionArtifactsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINGESTIONSESSIONARTIFACTSREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListIngestionSessionArtifactsRequest)
  })
_sym_db.RegisterMessage(ListIngestionSessionArtifactsRequest)

ListIngestionSessionArtifactsResponse = _reflection.GeneratedProtocolMessageType('ListIngestionSessionArtifactsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTINGESTIONSESSIONARTIFACTSRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListIngestionSessionArtifactsResponse)
  })
_sym_db.RegisterMessage(ListIngestionSessionArtifactsResponse)

ListIngestionSessionsRequest = _reflection.GeneratedProtocolMessageType('ListIngestionSessionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINGESTIONSESSIONSREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListIngestionSessionsRequest)
  })
_sym_db.RegisterMessage(ListIngestionSessionsRequest)

ListIngestionSessionsResponse = _reflection.GeneratedProtocolMessageType('ListIngestionSessionsResponse', (_message.Message,), {

  'IngestionSession' : _reflection.GeneratedProtocolMessageType('IngestionSession', (_message.Message,), {
    'DESCRIPTOR' : _LISTINGESTIONSESSIONSRESPONSE_INGESTIONSESSION,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ListIngestionSessionsResponse.IngestionSession)
    })
  ,
  'DESCRIPTOR' : _LISTINGESTIONSESSIONSRESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListIngestionSessionsResponse)
  })
_sym_db.RegisterMessage(ListIngestionSessionsResponse)
_sym_db.RegisterMessage(ListIngestionSessionsResponse.IngestionSession)

ApplyPerspectiveRequest = _reflection.GeneratedProtocolMessageType('ApplyPerspectiveRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLYPERSPECTIVEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyPerspectiveRequest)
  })
_sym_db.RegisterMessage(ApplyPerspectiveRequest)

ApplyPerspectiveResponse = _reflection.GeneratedProtocolMessageType('ApplyPerspectiveResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPLYPERSPECTIVERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ApplyPerspectiveResponse)
  })
_sym_db.RegisterMessage(ApplyPerspectiveResponse)

RetractPerspectiveRequest = _reflection.GeneratedProtocolMessageType('RetractPerspectiveRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTPERSPECTIVEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractPerspectiveRequest)
  })
_sym_db.RegisterMessage(RetractPerspectiveRequest)

RetractPerspectiveResponse = _reflection.GeneratedProtocolMessageType('RetractPerspectiveResponse', (_message.Message,), {
  'DESCRIPTOR' : _RETRACTPERSPECTIVERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.RetractPerspectiveResponse)
  })
_sym_db.RegisterMessage(RetractPerspectiveResponse)

ListPerspectiveRequest = _reflection.GeneratedProtocolMessageType('ListPerspectiveRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPERSPECTIVEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListPerspectiveRequest)
  })
_sym_db.RegisterMessage(ListPerspectiveRequest)

ListPerspectiveResponse = _reflection.GeneratedProtocolMessageType('ListPerspectiveResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPERSPECTIVERESPONSE,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ListPerspectiveResponse)
  })
_sym_db.RegisterMessage(ListPerspectiveResponse)

EventMetadata = _reflection.GeneratedProtocolMessageType('EventMetadata', (_message.Message,), {
  'DESCRIPTOR' : _EVENTMETADATA,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.EventMetadata)
  })
_sym_db.RegisterMessage(EventMetadata)

PublishResourceFailedEvent = _reflection.GeneratedProtocolMessageType('PublishResourceFailedEvent', (_message.Message,), {
  'DESCRIPTOR' : _PUBLISHRESOURCEFAILEDEVENT,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PublishResourceFailedEvent)
  })
_sym_db.RegisterMessage(PublishResourceFailedEvent)

PublishResourceCompleteEvent = _reflection.GeneratedProtocolMessageType('PublishResourceCompleteEvent', (_message.Message,), {
  'DESCRIPTOR' : _PUBLISHRESOURCECOMPLETEEVENT,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.PublishResourceCompleteEvent)
  })
_sym_db.RegisterMessage(PublishResourceCompleteEvent)

ExecuteIngestionSessionCompleteEvent = _reflection.GeneratedProtocolMessageType('ExecuteIngestionSessionCompleteEvent', (_message.Message,), {
  'DESCRIPTOR' : _EXECUTEINGESTIONSESSIONCOMPLETEEVENT,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ExecuteIngestionSessionCompleteEvent)
  })
_sym_db.RegisterMessage(ExecuteIngestionSessionCompleteEvent)

ExecuteIngestionSessionFailedEvent = _reflection.GeneratedProtocolMessageType('ExecuteIngestionSessionFailedEvent', (_message.Message,), {
  'DESCRIPTOR' : _EXECUTEINGESTIONSESSIONFAILEDEVENT,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ExecuteIngestionSessionFailedEvent)
  })
_sym_db.RegisterMessage(ExecuteIngestionSessionFailedEvent)

IngestionSessionCreatedEvent = _reflection.GeneratedProtocolMessageType('IngestionSessionCreatedEvent', (_message.Message,), {
  'DESCRIPTOR' : _INGESTIONSESSIONCREATEDEVENT,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.IngestionSessionCreatedEvent)
  })
_sym_db.RegisterMessage(IngestionSessionCreatedEvent)

ResolveUniverseRequest = _reflection.GeneratedProtocolMessageType('ResolveUniverseRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESOLVEUNIVERSEREQUEST,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveUniverseRequest)
  })
_sym_db.RegisterMessage(ResolveUniverseRequest)

ResolveUniverseGroupEntry = _reflection.GeneratedProtocolMessageType('ResolveUniverseGroupEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOLVEUNIVERSEGROUPENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveUniverseGroupEntry)
  })
_sym_db.RegisterMessage(ResolveUniverseGroupEntry)

ResolveUniverseEntry = _reflection.GeneratedProtocolMessageType('ResolveUniverseEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOLVEUNIVERSEENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ResolveUniverseEntry)
  })
_sym_db.RegisterMessage(ResolveUniverseEntry)

ErrorEntry = _reflection.GeneratedProtocolMessageType('ErrorEntry', (_message.Message,), {

  'ContextEntry' : _reflection.GeneratedProtocolMessageType('ContextEntry', (_message.Message,), {
    'DESCRIPTOR' : _ERRORENTRY_CONTEXTENTRY,
    '__module__' : 'bakplane.bakplane_pb2'
    # @@protoc_insertion_point(class_scope:io.openaristos.ErrorEntry.ContextEntry)
    })
  ,
  'DESCRIPTOR' : _ERRORENTRY,
  '__module__' : 'bakplane.bakplane_pb2'
  # @@protoc_insertion_point(class_scope:io.openaristos.ErrorEntry)
  })
_sym_db.RegisterMessage(ErrorEntry)
_sym_db.RegisterMessage(ErrorEntry.ContextEntry)


DESCRIPTOR._options = None
_TIMESTAMPRANGE.fields_by_name['start_dt']._options = None
_TIMESTAMPRANGE.fields_by_name['end_dt']._options = None
_UNIVERSE.fields_by_name['definition']._options = None
_PERSPECTIVE.fields_by_name['name']._options = None
_PERSPECTIVE.fields_by_name['code']._options = None
_PERSPECTIVE.fields_by_name['weight']._options = None
_PERSPECTIVEPERMANENTATTRIBUTE.fields_by_name['entity_type_code']._options = None
_PERSPECTIVEPERMANENTATTRIBUTE.fields_by_name['attribute_code']._options = None
_MIGRATIONREQUEST.fields_by_name['code']._options = None
_MIGRATIONREQUEST.fields_by_name['warehouse_kind']._options = None
_MIGRATIONREQUEST.fields_by_name['direction']._options = None
_MIGRATIONSTATUSREQUEST.fields_by_name['code']._options = None
_MIGRATIONSTATUSREQUEST.fields_by_name['warehouse_kind']._options = None
_RESOLVERESOURCEREQUEST_POINTERSENTRY._options = None
_RESOLVERESOURCEREQUEST.fields_by_name['universe']._options = None
_RESOLVERESOURCEINTENTRESPONSE_CONTEXTENTRY._options = None
_CONFIGURATIONENTRY.fields_by_name['key']._options = None
_CONFIGURATIONENTRY.fields_by_name['value']._options = None
_APPLYLANDINGZONEREQUEST.fields_by_name['code']._options = None
_RETRACTLANDINGZONEREQUEST.fields_by_name['code']._options = None
_LISTLANDINGZONEREQUEST.fields_by_name['query']._options = None
_ENABLELANDINGZONEREQUEST.fields_by_name['code']._options = None
_DISABLELANDINGZONEREQUEST.fields_by_name['code']._options = None
_SCHEMAATTRIBUTE.fields_by_name['name']._options = None
_SCHEMAATTRIBUTE.fields_by_name['data_type']._options = None
_SCHEMAATTRIBUTE.fields_by_name['mode']._options = None
_SCHEMAATTRIBUTE.fields_by_name['mapping_type']._options = None
_SCHEMA.fields_by_name['code']._options = None
_SCHEMA.fields_by_name['attributes']._options = None
_ENTITYTYPE.fields_by_name['code']._options = None
_ENTITYTYPE.fields_by_name['schema']._options = None
_RELATIONSHIPTYPE.fields_by_name['code']._options = None
_RELATIONSHIPTYPE.fields_by_name['source_entity_type']._options = None
_RELATIONSHIPTYPE.fields_by_name['target_entity_type']._options = None
_APPLYONTOLOGYREQUEST.fields_by_name['code']._options = None
_APPLYONTOLOGYREQUEST.fields_by_name['entities']._options = None
_RETRACTONTOLOGYREQUEST.fields_by_name['mastering_engine_code']._options = None
_RETRACTONTOLOGYREQUEST.fields_by_name['code']._options = None
_LISTONTOLOGYREQUEST.fields_by_name['query']._options = None
_EXECUTIONENGINE.fields_by_name['code']._options = None
_APPLYEXECUTIONENGINEREQUEST.fields_by_name['code']._options = None
_RETRACTEXECUTIONENGINEREQUEST.fields_by_name['code']._options = None
_ENABLEEXECUTIONENGINEREQUEST.fields_by_name['code']._options = None
_DISABLEEXECUTIONENGINEREQUEST.fields_by_name['code']._options = None
_WAREHOUSE.fields_by_name['code']._options = None
_APPLYWAREHOUSEREQUEST.fields_by_name['code']._options = None
_RETRACTWAREHOUSEREQUEST.fields_by_name['code']._options = None
_ENABLEWAREHOUSEREQUEST.fields_by_name['code']._options = None
_DISABLEWAREHOUSEREQUEST.fields_by_name['code']._options = None
_APPLYMASTERINGENGINEREQUEST.fields_by_name['code']._options = None
_RETRACTMASTERINGENGINEREQUEST.fields_by_name['code']._options = None
_RESOURCESTORE.fields_by_name['code']._options = None
_RESOURCEPOINTERENTRY.fields_by_name['code']._options = None
_RESOURCEPOINTERENTRY.fields_by_name['entity_type']._options = None
_RESOURCE.fields_by_name['code']._options = None
_RESOURCE.fields_by_name['perspective_code']._options = None
_RESOURCE.fields_by_name['schema']._options = None
_RESOURCE.fields_by_name['migrations']._options = None
_ADAPTER.fields_by_name['code']._options = None
_APPLYADAPTERREQUEST.fields_by_name['name']._options = None
_APPLYADAPTERREQUEST.fields_by_name['code']._options = None
_APPLYADAPTERREQUEST.fields_by_name['author']._options = None
_APPLYADAPTERREQUEST.fields_by_name['provider']._options = None
_APPLYADAPTERREQUEST.fields_by_name['execution_profile']._options = None
_APPLYADAPTERREQUEST.fields_by_name['resources']._options = None
_RETRACTADAPTERREQUEST.fields_by_name['code']._options = None
_ENABLEADAPTERREQUEST.fields_by_name['code']._options = None
_DISABLEADAPTERREQUEST.fields_by_name['code']._options = None
_APPLYRESOURCEREQUEST.fields_by_name['resource']._options = None
_RETRACTRESOURCEREQUEST.fields_by_name['code']._options = None
_ENABLERESOURCEREQUEST.fields_by_name['code']._options = None
_DISABLERESOURCEREQUEST.fields_by_name['code']._options = None
_ADAPTERPARAMETERS.fields_by_name['data_type']._options = None
_ADAPTEREXECUTIONCONTEXTENTRY.fields_by_name['data_type']._options = None
_ADAPTEREXECUTIONPROFILE.fields_by_name['engine_kind']._options = None
_ADAPTEREXECUTIONPROFILE.fields_by_name['context']._options = None
_CREATEINGESTIONSESSIONREQUEST_PARAMETERSENTRY._options = None
_CREATEINGESTIONSESSIONREQUEST.fields_by_name['adapter']._options = None
_CREATEINGESTIONSESSIONREQUEST.fields_by_name['landing_zone']._options = None
_CREATEINGESTIONSESSIONREQUEST.fields_by_name['mastering_engine']._options = None
_CREATEINGESTIONSESSIONREQUEST.fields_by_name['warehouses']._options = None
_EXECUTEINGESTIONSESSIONREQUEST_PARAMETERSENTRY._options = None
_INGESTIONSESSIONSTATUSRESPONSE.fields_by_name['status']._options = None
_MASTERENTITY_DESCRIPTORSENTRY._options = None
_PUBLISHINGESTIONSESSIONSTATUSRESPONSE.fields_by_name['status']._options = None
_RESOLVERESOURCEPATHREQUEST_CONTEXTENTRY._options = None
_RESOLVERESOURCEPATHRESPONSE_CONTEXTENTRY._options = None
_APPLYPERSPECTIVEREQUEST.fields_by_name['perspective']._options = None
_RESOLVEUNIVERSEENTRY.fields_by_name['data_type']._options = None
_ERRORENTRY_CONTEXTENTRY._options = None

_BAKPLANE = _descriptor.ServiceDescriptor(
  name='Bakplane',
  full_name='io.openaristos.Bakplane',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=15815,
  serialized_end=21468,
  methods=[
  _descriptor.MethodDescriptor(
    name='ApplyAdapter',
    full_name='io.openaristos.Bakplane.ApplyAdapter',
    index=0,
    containing_service=None,
    input_type=_APPLYADAPTERREQUEST,
    output_type=_APPLYADAPTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractAdapter',
    full_name='io.openaristos.Bakplane.RetractAdapter',
    index=1,
    containing_service=None,
    input_type=_RETRACTADAPTERREQUEST,
    output_type=_RETRACTADAPTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListAdapter',
    full_name='io.openaristos.Bakplane.ListAdapter',
    index=2,
    containing_service=None,
    input_type=_LISTADAPTERREQUEST,
    output_type=_LISTADAPTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='EnableAdapter',
    full_name='io.openaristos.Bakplane.EnableAdapter',
    index=3,
    containing_service=None,
    input_type=_ENABLEADAPTERREQUEST,
    output_type=_ENABLEADAPTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DisableAdapter',
    full_name='io.openaristos.Bakplane.DisableAdapter',
    index=4,
    containing_service=None,
    input_type=_DISABLEADAPTERREQUEST,
    output_type=_DISABLEADAPTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ApplyResource',
    full_name='io.openaristos.Bakplane.ApplyResource',
    index=5,
    containing_service=None,
    input_type=_APPLYRESOURCEREQUEST,
    output_type=_APPLYRESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractResource',
    full_name='io.openaristos.Bakplane.RetractResource',
    index=6,
    containing_service=None,
    input_type=_RETRACTRESOURCEREQUEST,
    output_type=_RETRACTRESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListResources',
    full_name='io.openaristos.Bakplane.ListResources',
    index=7,
    containing_service=None,
    input_type=_LISTRESOURCEREQUEST,
    output_type=_LISTRESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='EnableResource',
    full_name='io.openaristos.Bakplane.EnableResource',
    index=8,
    containing_service=None,
    input_type=_ENABLERESOURCEREQUEST,
    output_type=_ENABLERESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DisableResource',
    full_name='io.openaristos.Bakplane.DisableResource',
    index=9,
    containing_service=None,
    input_type=_DISABLERESOURCEREQUEST,
    output_type=_DISABLERESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='MigrateResource',
    full_name='io.openaristos.Bakplane.MigrateResource',
    index=10,
    containing_service=None,
    input_type=_MIGRATIONREQUEST,
    output_type=_MIGRATIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResourceMigrationStatus',
    full_name='io.openaristos.Bakplane.ResourceMigrationStatus',
    index=11,
    containing_service=None,
    input_type=_MIGRATIONSTATUSREQUEST,
    output_type=_MIGRATIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResolveResource',
    full_name='io.openaristos.Bakplane.ResolveResource',
    index=12,
    containing_service=None,
    input_type=_RESOLVERESOURCEREQUEST,
    output_type=_RESOLVERESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResolveResourceIntent',
    full_name='io.openaristos.Bakplane.ResolveResourceIntent',
    index=13,
    containing_service=None,
    input_type=_RESOLVERESOURCEREQUEST,
    output_type=_RESOLVERESOURCEINTENTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResolveUniverse',
    full_name='io.openaristos.Bakplane.ResolveUniverse',
    index=14,
    containing_service=None,
    input_type=_RESOLVEUNIVERSEREQUEST,
    output_type=_RESOLVEUNIVERSEGROUPENTRY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ApplyLandingZone',
    full_name='io.openaristos.Bakplane.ApplyLandingZone',
    index=15,
    containing_service=None,
    input_type=_APPLYLANDINGZONEREQUEST,
    output_type=_APPLYLANDINGZONERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractLandingZone',
    full_name='io.openaristos.Bakplane.RetractLandingZone',
    index=16,
    containing_service=None,
    input_type=_RETRACTLANDINGZONEREQUEST,
    output_type=_RETRACTLANDINGZONERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListLandingZone',
    full_name='io.openaristos.Bakplane.ListLandingZone',
    index=17,
    containing_service=None,
    input_type=_LISTLANDINGZONEREQUEST,
    output_type=_LISTLANDINGZONERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='EnableLandingZone',
    full_name='io.openaristos.Bakplane.EnableLandingZone',
    index=18,
    containing_service=None,
    input_type=_ENABLELANDINGZONEREQUEST,
    output_type=_ENABLELANDINGZONERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DisableLandingZone',
    full_name='io.openaristos.Bakplane.DisableLandingZone',
    index=19,
    containing_service=None,
    input_type=_DISABLELANDINGZONEREQUEST,
    output_type=_DISABLELANDINGZONERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ApplyOntology',
    full_name='io.openaristos.Bakplane.ApplyOntology',
    index=20,
    containing_service=None,
    input_type=_APPLYONTOLOGYREQUEST,
    output_type=_APPLYONTOLOGYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractOntology',
    full_name='io.openaristos.Bakplane.RetractOntology',
    index=21,
    containing_service=None,
    input_type=_RETRACTONTOLOGYREQUEST,
    output_type=_RETRACTONTOLOGYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListOntology',
    full_name='io.openaristos.Bakplane.ListOntology',
    index=22,
    containing_service=None,
    input_type=_LISTONTOLOGYREQUEST,
    output_type=_LISTONTOLOGYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ApplyExecutionEngine',
    full_name='io.openaristos.Bakplane.ApplyExecutionEngine',
    index=23,
    containing_service=None,
    input_type=_APPLYEXECUTIONENGINEREQUEST,
    output_type=_APPLYEXECUTIONENGINERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractExecutionEngine',
    full_name='io.openaristos.Bakplane.RetractExecutionEngine',
    index=24,
    containing_service=None,
    input_type=_RETRACTEXECUTIONENGINEREQUEST,
    output_type=_RETRACTEXECUTIONENGINERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListExecutionEngine',
    full_name='io.openaristos.Bakplane.ListExecutionEngine',
    index=25,
    containing_service=None,
    input_type=_LISTEXECUTIONENGINEREQUEST,
    output_type=_LISTEXECUTIONENGINERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='EnableExecutionEngine',
    full_name='io.openaristos.Bakplane.EnableExecutionEngine',
    index=26,
    containing_service=None,
    input_type=_ENABLEEXECUTIONENGINEREQUEST,
    output_type=_ENABLEEXECUTIONENGINERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DisableExecutionEngine',
    full_name='io.openaristos.Bakplane.DisableExecutionEngine',
    index=27,
    containing_service=None,
    input_type=_DISABLEEXECUTIONENGINEREQUEST,
    output_type=_DISABLEEXECUTIONENGINERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ApplyWarehouse',
    full_name='io.openaristos.Bakplane.ApplyWarehouse',
    index=28,
    containing_service=None,
    input_type=_APPLYWAREHOUSEREQUEST,
    output_type=_APPLYWAREHOUSERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractWarehouse',
    full_name='io.openaristos.Bakplane.RetractWarehouse',
    index=29,
    containing_service=None,
    input_type=_RETRACTWAREHOUSEREQUEST,
    output_type=_RETRACTWAREHOUSERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListWarehouse',
    full_name='io.openaristos.Bakplane.ListWarehouse',
    index=30,
    containing_service=None,
    input_type=_LISTWAREHOUSEREQUEST,
    output_type=_LISTWAREHOUSERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='EnableWarehouse',
    full_name='io.openaristos.Bakplane.EnableWarehouse',
    index=31,
    containing_service=None,
    input_type=_ENABLEWAREHOUSEREQUEST,
    output_type=_ENABLEWAREHOUSERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DisableWarehouse',
    full_name='io.openaristos.Bakplane.DisableWarehouse',
    index=32,
    containing_service=None,
    input_type=_DISABLEWAREHOUSEREQUEST,
    output_type=_DISABLEWAREHOUSERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ApplyMasteringEngine',
    full_name='io.openaristos.Bakplane.ApplyMasteringEngine',
    index=33,
    containing_service=None,
    input_type=_APPLYMASTERINGENGINEREQUEST,
    output_type=_APPLYMASTERINGENGINERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractMasteringEngine',
    full_name='io.openaristos.Bakplane.RetractMasteringEngine',
    index=34,
    containing_service=None,
    input_type=_RETRACTMASTERINGENGINEREQUEST,
    output_type=_RETRACTMASTERINGENGINERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateIngestionSession',
    full_name='io.openaristos.Bakplane.CreateIngestionSession',
    index=35,
    containing_service=None,
    input_type=_CREATEINGESTIONSESSIONREQUEST,
    output_type=_CREATEINGESTIONSESSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ExecuteIngestionSession',
    full_name='io.openaristos.Bakplane.ExecuteIngestionSession',
    index=36,
    containing_service=None,
    input_type=_EXECUTEINGESTIONSESSIONREQUEST,
    output_type=_INGESTIONSESSIONSTATUSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetIngestionSession',
    full_name='io.openaristos.Bakplane.GetIngestionSession',
    index=37,
    containing_service=None,
    input_type=_GETINGESTIONSESSIONREQUEST,
    output_type=_INGESTIONSESSIONSTATUSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='MasterIngestionSession',
    full_name='io.openaristos.Bakplane.MasterIngestionSession',
    index=38,
    containing_service=None,
    input_type=_MASTERINGESTIONSESSIONREQUEST,
    output_type=_MASTERINGESTIONSESSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='HaltIncrementalMastering',
    full_name='io.openaristos.Bakplane.HaltIncrementalMastering',
    index=39,
    containing_service=None,
    input_type=_HALTINCREMENTALMASTERINGREQUEST,
    output_type=_HALTINCREMENTALMASTERINGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResumeIncrementalMastering',
    full_name='io.openaristos.Bakplane.ResumeIncrementalMastering',
    index=40,
    containing_service=None,
    input_type=_RESUMEINCREMENTALMASTERINGREQUEST,
    output_type=_RESUMEINCREMENTALMASTERINGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PublishIngestionSessionResource',
    full_name='io.openaristos.Bakplane.PublishIngestionSessionResource',
    index=41,
    containing_service=None,
    input_type=_PUBLISHINGESTIONSESSIONRESOURCEREQUEST,
    output_type=_PUBLISHINGESTIONSESSIONRESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PublishIngestionSessionResourceStatus',
    full_name='io.openaristos.Bakplane.PublishIngestionSessionResourceStatus',
    index=42,
    containing_service=None,
    input_type=_PUBLISHINGESTIONSESSIONRESOURCESTATUSREQUEST,
    output_type=_PUBLISHINGESTIONSESSIONSTATUSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractPublishingSession',
    full_name='io.openaristos.Bakplane.RetractPublishingSession',
    index=43,
    containing_service=None,
    input_type=_RETRACTPUBLISHINGSESSIONREQUEST,
    output_type=_RETRACTPUBLISHINGSESSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractMasteringSession',
    full_name='io.openaristos.Bakplane.RetractMasteringSession',
    index=44,
    containing_service=None,
    input_type=_RETRACTMASTERINGSESSIONREQUEST,
    output_type=_RETRACTMASTERINGSESSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResolveResourcePath',
    full_name='io.openaristos.Bakplane.ResolveResourcePath',
    index=45,
    containing_service=None,
    input_type=_RESOLVERESOURCEPATHREQUEST,
    output_type=_RESOLVERESOURCEPATHRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListIngestionSessionArtifacts',
    full_name='io.openaristos.Bakplane.ListIngestionSessionArtifacts',
    index=46,
    containing_service=None,
    input_type=_LISTINGESTIONSESSIONARTIFACTSREQUEST,
    output_type=_LISTINGESTIONSESSIONARTIFACTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListIngestionSessions',
    full_name='io.openaristos.Bakplane.ListIngestionSessions',
    index=47,
    containing_service=None,
    input_type=_LISTINGESTIONSESSIONSREQUEST,
    output_type=_LISTINGESTIONSESSIONSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SessionPS',
    full_name='io.openaristos.Bakplane.SessionPS',
    index=48,
    containing_service=None,
    input_type=_PSREQUEST,
    output_type=_PSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ApplyPerspective',
    full_name='io.openaristos.Bakplane.ApplyPerspective',
    index=49,
    containing_service=None,
    input_type=_APPLYPERSPECTIVEREQUEST,
    output_type=_APPLYPERSPECTIVERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RetractPerspective',
    full_name='io.openaristos.Bakplane.RetractPerspective',
    index=50,
    containing_service=None,
    input_type=_RETRACTPERSPECTIVEREQUEST,
    output_type=_RETRACTPERSPECTIVERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListPerspective',
    full_name='io.openaristos.Bakplane.ListPerspective',
    index=51,
    containing_service=None,
    input_type=_LISTPERSPECTIVEREQUEST,
    output_type=_LISTPERSPECTIVERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BAKPLANE)

DESCRIPTOR.services_by_name['Bakplane'] = _BAKPLANE

# @@protoc_insertion_point(module_scope)
